# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""That which displays a one-way connection between two places.

An arrow connects two spots, the origin and the destination, and it
points from the origin to the destination, regardless of where on the
screen they are at the moment.

"""

from collections import defaultdict
from math import atan, cos, pi, sin
from operator import itemgetter
from typing import Iterator, Optional, Tuple

import numpy as np
from kivy.clock import Clock, mainthread
from kivy.core.text import Label
from kivy.graphics import Color, InstructionGroup, Quad, Rectangle, Translate
from kivy.graphics.fbo import Fbo
from kivy.properties import ListProperty, NumericProperty
from kivy.uix.widget import Widget

try:
	from kivy.garden.collider import Collide2DPoly
except (KeyError, ImportError):
	try:
		from kivy_garden.collider import Collide2DPoly
	except ImportError:
		from ..collide import Collide2DPoly

fortyfive = pi / 4
cos45 = cos(fortyfive)
sin45 = sin(fortyfive)


def up_and_down(orig, dest, taillen):
	if orig.center_y == dest.center_y:
		raise ValueError("Can't draw an arrow at a point")
	flipped = orig.center_y > dest.center_y
	if flipped:
		orig, dest = dest, orig
	x = int(orig.center_x)
	dy = int(dest.y)
	oy = int(orig.top)
	if flipped:
		oy, dy = dy, oy
	off1 = cos45 * taillen
	off2 = sin45 * taillen
	x0 = x
	y0 = oy
	endx = x
	endy = dy
	x1 = endx - off1
	x2 = endx + off1
	y1 = y2 = endy - off2 if oy < dy else endy + off2
	return [x0, y0, endx, endy], [x1, y1, endx, endy, x2, y2]


def left_and_right(orig, dest, taillen):
	if orig.center_x == dest.center_x:
		raise ValueError("Can't draw an arrow at a point")
	flipped = orig.center_x > dest.center_x
	if flipped:
		orig, dest = dest, orig
	y = int(orig.center_y)
	dx = int(dest.x)
	ox = int(orig.right)
	if flipped:
		ox, dx = dx, ox
	off1 = cos45 * taillen
	off2 = sin45 * taillen
	x0 = ox
	y0 = y
	endx = dx
	endy = y
	y1 = endy - off1
	y2 = endy + off1
	x1 = x2 = endx - off2 if ox < dx else endx + off2
	return [x0, y0, endx, endy], [x1, y1, endx, endy, x2, y2]


def _get_points_first_part(orig, dest, taillen):
	ox, oy = orig.center
	ow, oh = orig.size
	dx, dy = dest.center
	dw, dh = dest.size
	# Flip the endpoints around so that the arrow faces up and right.
	# We'll flip it back at the end as needed.
	xco = 1 if ox < dx else -1
	yco = 1 if oy < dy else -1
	ox *= xco
	dx *= xco
	oy *= yco
	dy *= yco
	# Nudge my endpoints so that I start and end at the edge of
	# a Spot.
	if dy - oy > dx - ox:
		topy = dy - dh / 2
		boty = oy + oh / 2
		leftx = ox
		rightx = dx
	else:
		leftx = ox + ow / 2
		rightx = dx - dw / 2
		topy = dy
		boty = oy
	# Degenerate cases.
	# Also, these need to be handled specially to avoid
	# division by zero later on.
	if rightx == leftx:
		return up_and_down(orig, dest, taillen)
	if topy == boty:
		return left_and_right(orig, dest, taillen)
	return ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty


def get_points_multi(args):
	"""Return a dictionary mapping (orig, dest) to pairs of point lists for arrows

	Takes an iterable of (orig, dest, taillen) where orig and dest are Spot instances

	taillen is an integer specifying how long the arrowhead should be.

	"""
	ret = {}
	keys = []
	topys = []
	botys = []
	leftxs = []
	rightxs = []
	taillens = []
	xcos = []
	ycos = []
	for orig, dest, taillen in args:
		try:
			p1 = _get_points_first_part(orig, dest, taillen)
		except ValueError:
			p1 = 2, 2, 2, 2, 1, 0, 1, 1, 0, 1
		if len(p1) == 2:
			ret[orig.name, dest.name] = p1
			continue
		ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty = p1
		keys.append((orig.name, dest.name))
		leftxs.append(leftx)
		rightxs.append(rightx)
		topys.append(topy)
		botys.append(boty)
		taillens.append(taillen)
		xcos.append(xco)
		ycos.append(yco)
	topys = np.array(topys)
	botys = np.array(botys)
	rightxs = np.array(rightxs)
	leftxs = np.array(leftxs)
	rises = np.subtract(topys, botys)
	runs = np.subtract(rightxs, leftxs)
	slopes = np.divide(rises, runs)
	unslopes = np.divide(runs, rises)
	start_thetas = np.arctan(slopes)
	end_thetas = np.arctan(unslopes)
	top_thetas = np.subtract(start_thetas, fortyfive)
	bot_thetas = np.subtract(pi - fortyfive, end_thetas)
	topsins = np.sin(top_thetas)
	topcoss = np.cos(top_thetas)
	botsins = np.sin(bot_thetas)
	botcoss = np.cos(bot_thetas)
	taillens = np.array(taillens)
	xoff1s = np.multiply(topcoss, taillens)
	yoff1s = np.multiply(topsins, taillens)
	xoff2s = np.multiply(botcoss, taillens)
	yoff2s = np.multiply(botsins, taillens)
	xcos = np.array(xcos)
	ycos = np.array(ycos)
	x1s = np.multiply(np.subtract(rightxs, xoff1s), xcos)
	x2s = np.multiply(np.subtract(rightxs, xoff2s), xcos)
	y1s = np.multiply(np.subtract(topys, yoff1s), ycos)
	y2s = np.multiply(np.subtract(topys, yoff2s), ycos)
	startxs = np.multiply(leftxs, xcos)
	startys = np.multiply(botys, ycos)
	endxs = np.multiply(rightxs, xcos)
	endys = np.multiply(topys, ycos)
	for key, startx, starty, endx, endy, x1, y1, endx, endy, x2, y2 in zip(
		keys, startxs, startys, endxs, endys, x1s, y1s, endxs, endys, x2s, y2s
	):
		ret[key] = ([startx, starty, endx, endy], [x1, y1, endx, endy, x2, y2])
	return ret


def get_points(orig, dest, taillen):
	"""Return a pair of lists of points for use making an arrow.

	The first list is the beginning and end point of the trunk of the arrow.

	The second list is the arrowhead.

	"""
	p1 = _get_points_first_part(orig, dest, taillen)
	if len(p1) == 2:
		return p1
	ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty = p1
	rise = topy - boty
	run = rightx - leftx
	slope = rise / run
	unslope = run / rise
	start_theta = atan(slope)
	end_theta = atan(unslope)

	# make the little wedge at the end so you can tell which way the
	# arrow's pointing, and flip it all back around to the way it was
	top_theta = start_theta - fortyfive
	bot_theta = pi - fortyfive - end_theta
	xoff1 = cos(top_theta) * taillen
	yoff1 = sin(top_theta) * taillen
	xoff2 = cos(bot_theta) * taillen
	yoff2 = sin(bot_theta) * taillen
	x1 = (rightx - xoff1) * xco
	x2 = (rightx - xoff2) * xco
	y1 = (topy - yoff1) * yco
	y2 = (topy - yoff2) * yco
	startx = leftx * xco
	starty = boty * yco
	endx = rightx * xco
	endy = topy * yco
	return [startx, starty, endx, endy], [x1, y1, endx, endy, x2, y2]


def get_quad_vertices(
	ox,
	oy,
	dx,
	dy,
	x1,
	y1,
	endx,
	endy,
	x2,
	y2,
	bgr,
	fgr,
	label_w=100,
	label_h=100,
):
	if dx > ox:
		if dy > oy:
			label_pos = x2 - label_w, dy - label_h
		elif dy < oy:
			label_pos = x1 - label_w, dy
		else:
			label_pos = x2 - label_w, y2 - label_h
	elif dx < ox:
		if dy < oy:
			label_pos = x2, y2
		elif dy > oy:
			label_pos = x1, y1 - label_h
		else:
			label_pos = x2, dy - label_h
	else:
		if dy > oy:
			label_pos = x2 - label_w, dy - label_h
		elif dy < oy:
			label_pos = x2, y2
		else:
			label_pos = dx, dy
	return {
		"shaft_bg": get_thin_rect_vertices(ox, oy, dx, dy, bgr),
		"shaft_fg": get_thin_rect_vertices(ox, oy, dx, dy, fgr),
		"left_head_bg": get_thin_rect_vertices(x1, y1, endx, endy, bgr),
		"right_head_bg": get_thin_rect_vertices(x2, y2, endx, endy, bgr),
		"left_head_fg": get_thin_rect_vertices(x1, y1, endx, endy, fgr),
		"right_head_fg": get_thin_rect_vertices(x2, y2, endx, endy, fgr),
		"label_pos": label_pos,
	}


eight0s = tuple([0] * 8)


class GraphArrow:
	@property
	def slope(self) -> float:
		"""Return a float of the increase in y divided by the increase in x,
		both from left to right.

		Returns ``None`` when vertical.

		"""
		orig = self.origin
		dest = self.destination
		ox = orig.x
		oy = orig.y
		dx = dest.x
		dy = dest.y
		if oy == dy:
			return 0.0
		elif ox == dx:
			return None
		else:
			rise = dy - oy
			run = dx - ox
			return rise / run

	@property
	def y_intercept(self) -> Optional[float]:
		"""Return my Y-intercept.

		I probably don't really hit the left edge of the window, but
		this is where I would, if I were long enough.

		"""
		orig = self.origin
		dest = self.destination
		(ox, oy) = orig.pos
		(dx, dy) = dest.pos
		denominator = dx - ox
		x_numerator = (dy - oy) * ox
		y_numerator = denominator * oy
		return (y_numerator - x_numerator), denominator

	@property
	def reciprocal(self) -> Optional["GraphArrow"]:
		"""The arrow connecting the same spots in the opposite direction"""
		if self.destination.name not in self.board.pred_arrow:
			return
		if (
			self.origin.name
			not in self.board.pred_arrow[self.destination.name]
		):
			return
		return self.board.pred_arrow[self.destination.name][self.origin.name]

	@property
	def selected(self) -> bool:
		return self is self.board.app.selection

	@selected.setter
	def selected(self, b: bool):
		self.board.app.selection = self
		self.repoint(b)

	@property
	def character(self):
		return self.board.character

	def __init__(self, *, board, origin, destination):
		self.board = board
		self.origin = origin
		self.destination = destination

	def collide_point(self, x: float, y: float) -> bool:
		od = (self.origin.name, self.destination.name)
		if od not in self.board.arrow_plane._colliders_map:
			if od not in self.board.arrow_plane._instructions_map:
				shaft_points, head_points = get_points(
					self.origin,
					self.destination,
					self.board.arrow_plane.arrowhead_size,
				)
				r = self.board.arrow_plane.arrow_width / 2
				bg_points = get_quad_vertices(
					*shaft_points,
					*head_points,
					r * self.board.arrow_plane.bg_scale_unselected,
					r,
					0,
					0,
				)["shaft_bg"]
			else:
				bg_points = self.board.arrow_plane._instructions_map[od][
					"shaft_bg"
				].points
			self.board.arrow_plane._colliders_map[od] = Collide2DPoly(
				points=bg_points, cache=True
			)
		return (x, y) in self.board.arrow_plane._colliders_map[
			self.origin.name, self.destination.name
		]

	def pos_along(self, pct: float) -> Tuple[float, float]:
		"""Return coordinates for where a Pawn should be if it has travelled
		along ``pct`` of my length (between 0 and 1).

		Might get complex when I switch over to using beziers for
		arrows, but for now this is quite simple, using distance along
		a line segment.

		"""
		if pct < 0 or pct > 1:
			raise ValueError("Invalid portion")
		(ox, oy) = self.origin.center
		(dx, dy) = self.destination.center
		xdist = (dx - ox) * pct
		ydist = (dy - oy) * pct
		return ox + xdist, oy + ydist

	@mainthread
	def repoint(self, selected: bool = None) -> None:
		arrow_plane = self.board.arrow_plane
		fbo = arrow_plane._fbo
		fbo.bind()
		fbo.clear_buffer()
		shaft_points, head_points = get_points(
			self.origin, self.destination, arrow_plane.arrowhead_size
		)
		r = arrow_plane.arrow_width / 2
		if selected or self.selected:
			bg_scale = arrow_plane.bg_scale_selected
			bg_color = arrow_plane.bg_color_selected
			fg_color = arrow_plane.fg_color_selected
		else:
			bg_scale = arrow_plane.bg_scale_unselected
			bg_color = arrow_plane.bg_color_unselected
			fg_color = arrow_plane.fg_color_unselected
		plane = self.board.arrow_plane
		portal = self.board.character.portal[self.origin.name][
			self.destination.name
		]
		portal_text = str(portal.get(portal.get("_label_stat", None), ""))
		label_kwargs = dict(portal.get("label_kwargs", ()))
		if portal_text is not None:
			label_kwargs["text"] = portal_text
		try:
			label = self.board.arrow_plane._labels[self.origin.name][
				self.destination.name
			]
			label.text = portal_text
		except KeyError:
			label = self.board.arrow_plane.labels[self.origin.name][
				self.destination.name
			] = Label(**DEFAULT_ARROW_LABEL_KWARGS, **label_kwargs)
		if (
			self.origin.name,
			self.destination.name,
		) in plane._instructions_map:
			verts = get_quad_vertices(
				*shaft_points, *head_points, r * bg_scale, r, *label.render()
			)
			insts = self.board.arrow_plane._instructions_map[
				self.origin.name, self.destination.name
			]
			insts["color0"].rgba = bg_color
			insts["color1"].rgba = fg_color
			insts["shaft_bg"].points = verts["shaft_bg"]
			insts["left_head_bg"].points = verts["left_head_bg"]
			insts["right_head_bg"].points = verts["right_head_bg"]
			insts["shaft_fg"].points = verts["shaft_fg"]
			insts["left_head_fg"].points = verts["left_head_fg"]
			insts["right_head_fg"].points = verts["right_head_fg"]
			insts["label_rect"].pos = verts["label_pos"]
			insts["label_rect"].size = label.render()
			label.refresh()
			insts["label_rect"].texture = label.texture
			plane._colliders_map[self.origin.name, self.destination.name] = (
				Collide2DPoly(points=verts["shaft_bg"])
			)
		else:
			plane._instructions_map[
				self.origin.name, self.destination.name
			] = insts = get_instructions(
				*shaft_points,
				*head_points,
				bg_color,
				fg_color,
				*label.render(),
				label,
			)
			plane._colliders_map[
				self.origin.name, self.destination.name
			].points = Collide2DPoly(points=insts["shaft_bg"].points)
		myidx = plane._port_index[self.origin.name, self.destination.name]
		(ox, oy, dx, dy) = shaft_points
		plane._bot_left_corner_xs[myidx] = min((ox, dx))
		plane._bot_left_corner_ys[myidx] = min((oy, dy))
		plane._top_right_corner_xs[myidx] = max((ox, dx))
		plane._top_right_corner_ys[myidx] = max((oy, dy))
		fbo.release()
		fbo.ask_update()
		arrow_plane.canvas.ask_update()


def get_instructions(
	ox: float,
	oy: float,
	dx: float,
	dy: float,
	x1: float,
	y1: float,
	endx: float,
	endy: float,
	x2: float,
	y2: float,
	bgr: float,
	r: float,
	bg_color: Tuple[float, float, float, float],
	fg_color: Tuple[float, float, float, float],
	label_kwargs: dict = None,
	label: Label = None,
) -> dict:
	if label_kwargs is None:
		label_kwargs = {"text": "", **DEFAULT_ARROW_LABEL_KWARGS}
	else:
		label_kwargs = dict(label_kwargs)
		for k, v in DEFAULT_ARROW_LABEL_KWARGS.items():
			if k not in label_kwargs:
				label_kwargs[k] = v
	if label is None:
		label = Label(**label_kwargs)
	else:
		label.text = label_kwargs["text"]
	text_size = label.render()
	quadverts = get_quad_vertices(
		ox, oy, dx, dy, x1, y1, endx, endy, x2, y2, bgr, r, *text_size
	)
	label.refresh()
	return {
		"color0": Color(rgba=bg_color),
		"shaft_bg": Quad(points=quadverts["shaft_bg"]),
		"left_head_bg": Quad(points=quadverts["left_head_bg"]),
		"right_head_bg": Quad(points=quadverts["right_head_bg"]),
		"color1": Color(rgba=fg_color),
		"shaft_fg": Quad(points=quadverts["shaft_fg"]),
		"left_head_fg": Quad(points=quadverts["left_head_fg"]),
		"right_head_fg": Quad(points=quadverts["right_head_fg"]),
		"label_rect": Rectangle(
			pos=quadverts["label_pos"], size=text_size, texture=label.texture
		),
		"label": label,
	}


class GraphArrowWidget(Widget, GraphArrow):
	arrowhead_size = NumericProperty(10)
	arrow_width = NumericProperty(2)
	bg_scale = NumericProperty(5)
	bg_color = ListProperty([0.5, 0.5, 0.5, 0.5])
	fg_color = ListProperty([1.0, 1.0, 1.0, 1.0])

	def __init__(self, **kwargs):
		self._trigger_repoint = Clock.create_trigger(self.repoint)
		super().__init__(**kwargs)

	def on_origin(self, *args):
		if hasattr(self, "_origin_bind_uid"):
			self.unbind_uid(self._origin_bind_uid)
			del self._origin_bind_uid
		if not self.origin:
			return
		self._origin_bind_uid = self.origin.fbind("pos", self._trigger_repoint)

	def on_destination(self, *args):
		if hasattr(self, "_destination_bind_uid"):
			self.unbind_uid(self._destination_bind_uid)
			del self._destination_bind_uid
		if not self.destination:
			return
		self._destination_bind_uid = self.destination.fbind(
			"pos", self._trigger_repoint
		)

	def on_parent(self, *args):
		if not self.origin or not self.destination or not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		self.canvas.clear()
		if self.parent is None:
			return
		shaft_points, head_points = get_points(
			self.origin, self.destination, self.arrowhead_size
		)
		with self.canvas:
			r = self.arrow_width / 2
			self._instructions = get_instructions(
				*shaft_points,
				*head_points,
				r * self.bg_scale,
				r,
				self.bg_color,
				self.fg_color,
			)

	def repoint(self, *args):
		shaft_points, head_points = get_points(
			self.origin, self.destination, self.arrowhead_size
		)
		r = self.arrow_width / 2
		try:
			portal = self.board.character.portal[self.origin.name][
				self.destination.name
			]
			portal_text = str(portal.get(portal.get("_label_stat", None), ""))
		except (KeyError, AttributeError):
			portal_text = ""
		if hasattr(self, "_label"):
			label = self._label
			label.text = portal_text
		else:
			label = self._label = Label(text=portal_text)
		label_size = label.render()
		verts = get_quad_vertices(
			*shaft_points, *head_points, r * self.bg_scale, r, *label_size
		)
		insts = self._instructions
		insts["color0"].rgba = self.bg_color
		insts["color1"].rgba = self.fg_color
		insts["shaft_bg"].points = verts["shaft_bg"]
		insts["left_head_bg"].points = verts["left_head_bg"]
		insts["right_head_bg"].points = verts["right_head_bg"]
		insts["shaft_fg"].points = verts["shaft_fg"]
		insts["left_head_fg"].points = verts["left_head_fg"]
		insts["right_head_fg"].points = verts["right_head_fg"]
		insts["label_rect"].pos = verts["label_pos"]
		insts["label_rect"].size = label_size
		label.refresh()
		insts["label_rect"].texture = label.texture


class ArrowPlane(Widget):
	data = ListProperty()
	arrowhead_size = NumericProperty(10)
	arrow_width = NumericProperty(2)
	bg_scale_unselected = NumericProperty(4)
	bg_scale_selected = NumericProperty(5)
	bg_color_selected = ListProperty([0.0, 1.0, 1.0, 1.0])
	bg_color_unselected = ListProperty([0.5, 0.5, 0.5, 0.5])
	fg_color_unselected = ListProperty([1.0, 1.0, 1.0, 1.0])
	fg_color_selected = ListProperty([1.0, 1.0, 1.0, 1.0])

	def __init__(self, **kwargs):
		self._labels = defaultdict(lambda: defaultdict(lambda: None))
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self._redraw_bind_uid = self.fbind("data", self._trigger_redraw)
		self.bind(arrowhead_size=self._trigger_redraw)
		self._colliders_map = {}
		self._instructions_map = {}
		self._port_index = {}
		self._port_l = []
		self._bot_left_corner_ys = np.array([])
		self._bot_left_corner_xs = np.array([])
		self._top_right_corner_ys = np.array([])
		self._top_right_corner_xs = np.array([])
		super().__init__(**kwargs)

	def on_parent(self, *args):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		with self.canvas:
			self._fbo = Fbo(size=self.size)
			self._translate = Translate(x=self.x, y=self.y)
			self._rectangle = Rectangle(
				size=self.size, texture=self._fbo.texture
			)
		self.redraw()

	def have_arrow(self, orig, dest):
		return (orig, dest) in self._port_index

	def redraw(self, *args):
		if not hasattr(self, "_rectangle"):
			self._trigger_redraw()
			return
		fbo = self._fbo
		fbo.bind()
		fbo.clear()
		fbo.clear_buffer()
		add = fbo.add
		r = self.arrow_width // 2
		bg_scale_selected = self.bg_scale_selected
		bg_color_unselected = self.bg_color_unselected
		fg_color_unselected = self.fg_color_unselected
		taillen = self.arrowhead_size
		points_map = get_points_multi(
			(datum["origspot"], datum["destspot"], taillen)
			for datum in self.data
		)
		port_l = []
		bot_left_corner_xs = []
		bot_left_corner_ys = []
		top_right_corner_xs = []
		top_right_corner_ys = []
		port_index = self._port_index
		colliders_map = self._colliders_map
		oxs = []
		oys = []
		dxs = []
		dys = []
		for (ox, oy, dx, dy), _ in points_map.values():
			oxs.append(ox)
			oys.append(oy)
			dxs.append(dx)
			dys.append(dy)
		widths = np.abs(np.array(dxs) - np.array(oxs))
		heights = np.abs(np.array(dys) - np.array(oys))
		lengths = np.sqrt(np.square(widths) + np.square(heights))
		for length, (
			port,
			((ox, oy, dx, dy), (x1, y1, endx, endy, x2, y2)),
		), datum in zip(lengths, points_map.items(), self.data):
			if length < r:
				continue
			bgr = r * bg_scale_selected  # change for selectedness pls
			instructions = get_instructions(
				ox,
				oy,
				dx,
				dy,
				x1,
				y1,
				endx,
				endy,
				x2,
				y2,
				bgr,
				r,
				bg_color_unselected,
				fg_color_unselected,
				datum.get("label_kwargs", {}),
				self._labels[port[0]].get(port[1]),
			)
			instructions["group"] = grp = InstructionGroup()
			grp.add(instructions["color0"])
			grp.add(instructions["shaft_bg"])
			grp.add(instructions["left_head_bg"])
			grp.add(instructions["right_head_bg"])
			grp.add(instructions["color1"])
			grp.add(instructions["shaft_fg"])
			grp.add(instructions["left_head_fg"])
			grp.add(instructions["right_head_fg"])
			grp.add(instructions["label_rect"])
			add(grp)
			self._instructions_map[port] = instructions
			self._labels[port[0]][port[1]] = instructions["label"]
			if ox < dx:
				leftx = ox
				rightx = dx
			else:
				rightx = ox
				leftx = dx
			if oy < dy:
				boty = oy
				topy = dy
			else:
				boty = dy
				topy = oy
			port_index[port] = len(port_l)
			port_l.append(port)
			bot_left_corner_xs.append(leftx - bgr)
			bot_left_corner_ys.append(boty - bgr)
			top_right_corner_xs.append(rightx + bgr)
			top_right_corner_ys.append(topy + bgr)
			colliders_map[port] = Collide2DPoly(
				points=instructions["shaft_bg"].points
			)
		fbo.release()
		self.canvas.ask_update()
		self._port_l = port_l
		self._bot_left_corner_xs = np.array(bot_left_corner_xs)
		self._bot_left_corner_ys = np.array(bot_left_corner_ys)
		self._top_right_corner_xs = np.array(top_right_corner_xs)
		self._top_right_corner_ys = np.array(top_right_corner_ys)

	def add_new_portal(self, datum: dict) -> None:
		orig_spot = datum["origspot"]
		dest_spot = datum["destspot"]
		shaft_points, head_points = get_points(
			orig_spot, dest_spot, self.arrowhead_size
		)
		self.unbind_uid("data", self._redraw_bind_uid)
		self.data.append(datum)
		r = self.arrow_width / 2
		bgr = r * self.bg_scale_unselected
		instructions = self._instructions_map[
			orig_spot.name, dest_spot.name
		] = get_instructions(
			*shaft_points,
			*head_points,
			bgr,
			r,
			self.bg_color_unselected,
			self.fg_color_unselected,
			datum.get("label_kwargs", {}),
			self._labels[orig_spot.name][dest_spot.name],
		)
		instructions["group"] = grp = InstructionGroup()
		grp.add(instructions["color0"])
		grp.add(instructions["shaft_bg"])
		grp.add(instructions["left_head_bg"])
		grp.add(instructions["right_head_bg"])
		grp.add(instructions["color1"])
		grp.add(instructions["shaft_fg"])
		grp.add(instructions["left_head_fg"])
		grp.add(instructions["right_head_fg"])
		# grp.add(instructions["label"])
		self._labels[orig_spot.name][dest_spot.name] = instructions["label"]
		self._fbo.add(grp)
		ox, oy, dx, dy = shaft_points
		if ox < dx:
			left_x = ox
			right_x = dx
		else:
			right_x = ox
			left_x = dx
		if oy < dy:
			bot_y = oy
			top_y = dy
		else:
			bot_y = dy
			top_y = oy
		self._port_index[orig_spot.name, dest_spot.name] = len(self._port_l)
		self._port_l.append((orig_spot.name, dest_spot.name))
		self._bot_left_corner_xs = np.array(
			list(self._bot_left_corner_xs) + [left_x - bgr]
		)
		self._bot_left_corner_ys = np.array(
			list(self._bot_left_corner_ys) + [bot_y - bgr]
		)
		self._top_right_corner_xs = np.array(
			list(self._top_right_corner_xs) + [right_x + bgr]
		)
		self._top_right_corner_ys = np.array(
			list(self._top_right_corner_ys) + [top_y + bgr]
		)
		self._colliders_map[orig_spot.name, dest_spot.name] = Collide2DPoly(
			points=instructions["shaft_bg"].points
		)
		self.canvas.ask_update()
		self._redraw_bind_uid = self.fbind("data", self._trigger_redraw)

	def remove_edge(self, orig, dest):
		self._fbo.bind()
		self._fbo.clear_buffer()
		self._fbo.remove(self._instructions_map[orig, dest]["group"])
		index = self._port_index[orig, dest]
		for port in self._port_l[index:]:
			self._port_index[port] -= 1
		del self._port_index[orig, dest]
		del self._instructions_map[orig, dest]
		del self._colliders_map[orig, dest]
		del self._port_l[index]
		for arr in (
			"_bot_left_corner_ys",
			"_bot_left_corner_xs",
			"_top_right_corner_ys",
			"_top_right_corner_xs",
		):
			dat = list(getattr(self, arr))
			del dat[index]
			setattr(self, arr, np.array(dat))
		self._fbo.release()
		self.canvas.ask_update()

	def update_portal_label(self, orig, dest, text):
		rect = self._instructions_map[orig, dest]["label"]
		label = self._labels[orig][dest]
		label.text = text
		label.refresh()
		rect.texture = label.texture

	def iter_collided_edges(self, x: float, y: float) -> Iterator:
		x, y = map(float, (x, y))
		collider_map = self._colliders_map
		hits = (
			(self._bot_left_corner_xs <= x)
			& (self._bot_left_corner_ys <= y)
			& (x <= self._top_right_corner_xs)
			& (y <= self._top_right_corner_ys)
		)
		for port in map(
			itemgetter(0), filter(itemgetter(1), zip(self._port_l, hits))
		):
			if collider_map[port].collide_point(x, y):
				yield port

	def on_pos(self, *args):
		if not hasattr(self, "_translate"):
			return
		self._translate.x, self._translate.y = self.pos
		self.canvas.ask_update()

	def on_size(self, *args):
		if not hasattr(self, "_rectangle") or not hasattr(self, "_fbo"):
			return
		self._rectangle.size = self._fbo.size = self.size
		self.redraw()


def get_thin_rect_vertices(ox, oy, dx, dy, r):
	"""Given the starting point, ending point, and width, return a list of
	vertex coordinates at the corners of the line segment
	(really a thin rectangle).

	"""
	if ox < dx:
		leftx = ox
		rightx = dx
		xco = 1
	elif ox > dx:
		leftx = ox * -1
		rightx = dx * -1
		xco = -1
	else:
		return [ox - r, oy, ox + r, oy, ox + r, dy, ox - r, dy]
	if oy < dy:
		boty = oy
		topy = dy
		yco = 1
	elif oy > dy:
		boty = oy * -1
		topy = dy * -1
		yco = -1
	else:
		return [ox, oy - r, dx, oy - r, dx, oy + r, ox, oy + r]

	rise = topy - boty
	run = rightx - leftx
	theta = atan(rise / run)
	theta_prime = ninety - theta
	xoff = cos(theta_prime) * r
	yoff = sin(theta_prime) * r
	x1 = leftx + xoff
	y1 = boty - yoff
	x2 = rightx + xoff
	y2 = topy - yoff
	x3 = rightx - xoff
	y3 = topy + yoff
	x4 = leftx - xoff
	y4 = boty + yoff
	return [
		x1 * xco,
		y1 * yco,
		x2 * xco,
		y2 * yco,
		x3 * xco,
		y3 * yco,
		x4 * xco,
		y4 * yco,
	]


ninety = pi / 2
DEFAULT_ARROW_LABEL_KWARGS = {"font_size": 16, "bold": True}



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial

from kivy.logger import Logger
from kivy.properties import (
	NumericProperty,
	ObjectProperty,
	ReferenceListProperty,
	StringProperty,
)

from .kivygarden.texturestack import ImageStack
from .util import logwrap, store_kv

wraplog_Dummy = partial(logwrap, section="Dummy")


class Dummy(ImageStack):
	"""A widget that looks like the ones on the graph, which, when dragged
	onto the graph, creates one of them.

	"""

	_touch = ObjectProperty(None, allownone=True)
	name = StringProperty()
	prefix = StringProperty()
	num = NumericProperty()
	x_start = NumericProperty(0)
	y_start = NumericProperty(0)
	pos_start = ReferenceListProperty(x_start, y_start)
	x_down = NumericProperty(0)
	y_down = NumericProperty(0)
	pos_down = ReferenceListProperty(x_down, y_down)
	x_up = NumericProperty(0)
	y_up = NumericProperty(0)
	pos_up = ReferenceListProperty(x_up, y_up)
	x_center_up = NumericProperty(0)
	y_center_up = NumericProperty(0)
	center_up = ReferenceListProperty(x_center_up, y_center_up)
	right_up = NumericProperty(0)
	top_up = NumericProperty(0)

	@logwrap(section="Dummy")
	def on_paths(self, *args, **kwargs):
		super().on_paths(*args, **kwargs)
		Logger.debug("Dummy: {} got paths {}".format(self.name, self.paths))

	@logwrap(section="Dummy")
	def on_touch_down(self, touch):
		"""If hit, record my starting position, that I may return to it in
		``on_touch_up`` after creating a real :class:`graph.Spot` or
		:class:`graph.Pawn` instance.

		"""
		if not self.collide_point(*touch.pos):
			return False
		self.pos_start = self.pos
		self.pos_down = (self.x - touch.x, self.y - touch.y)
		touch.grab(self)
		self._touch = touch
		return True

	@logwrap(section="Dummy")
	def on_touch_move(self, touch):
		"""Follow the touch"""
		if touch is not self._touch:
			return False
		self.pos = (touch.x + self.x_down, touch.y + self.y_down)
		return True

	@logwrap(section="Dummy")
	def on_touch_up(self, touch):
		"""Return to ``pos_start``, but first, save my current ``pos`` into
		``pos_up``, so that the layout knows where to put the real
		:class:`graph.Spot` or :class:`graph.Pawn` instance.

		"""
		if touch is not self._touch:
			return False
		self.pos_up = self.pos
		self.pos = self.pos_start
		self._touch = None
		return True


store_kv(
	__name__,
	"""
<Dummy>:
	name: "".join((self.prefix, str(self.num)))
	x_center_up: self.x_up + self.width / 2
	y_center_up: self.y_up + self.height / 2
	right_up: self.x_up + self.width
	top_up: self.y_up + self.height
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Code that draws the box around a Pawn or Spot when it's selected"""

from collections import defaultdict
from functools import partial

from kivy.clock import Clock, triggered
from kivy.graphics import (
	Color,
	InstructionGroup,
	Line,
	PopMatrix,
	PushMatrix,
	Translate,
)
from kivy.properties import BooleanProperty, ListProperty, ObjectProperty
from kivy.uix.layout import Layout

from elide.imagestackproxy import ImageStackProxy

from .util import logwrap


def trigger(func):
	return triggered()(func)


class GraphPawnSpot(ImageStackProxy, Layout):
	"""The kind of ImageStack that represents a :class:`Thing` or
	:class:`Place`.

	"""

	board = ObjectProperty()
	engine = ObjectProperty()
	selected = BooleanProperty(False)
	linecolor = ListProperty()
	selected_outline_color = ListProperty([0, 1, 1, 1])
	unselected_outline_color = ListProperty([0, 0, 0, 0])
	use_boardspace = True

	def __init__(self, **kwargs):
		if "proxy" in kwargs:
			kwargs["name"] = kwargs["proxy"].name
		super().__init__(**kwargs)
		self.bind(pos=self._position)

	@logwrap(section="GraphPawnSpot")
	def on_touch_move(self, touch):
		"""If I'm being dragged, move to follow the touch."""
		if touch.grab_current is not self:
			return False
		self.center = touch.pos
		return True

	@logwrap(section="GraphPawnSpot")
	def finalize(self, initial=True):
		"""Call this after you've created all the PawnSpot you need and are ready to add them to the board."""
		if getattr(self, "_finalized", False):
			return
		if self.proxy is None or not hasattr(self.proxy, "name"):
			Clock.schedule_once(partial(self.finalize, initial=initial), 0)
			return
		if initial:
			self.name = self.proxy.name
			if "_image_paths" in self.proxy:
				try:
					self.paths = self.proxy["_image_paths"]
				except Exception as ex:
					if not (
						isinstance(ex.args[0], str)
						and ex.args[0].startswith("Unable to load image type")
					):
						raise ex
					self.paths = self.default_image_paths
			else:
				self.paths = self.proxy.setdefault(
					"_image_paths", self.default_image_paths
				)
			zeroes = [0] * len(self.paths)
			self.offxs = self.proxy.setdefault("_offxs", zeroes)
			self.offys = self.proxy.setdefault("_offys", zeroes)
			self.proxy.connect(self._trigger_pull_from_proxy)
			self.finalize_children(initial=True)
		self._push_image_paths_binding = self.fbind(
			"paths", self._trigger_push_image_paths
		)
		self._push_offxs_binding = self.fbind(
			"offxs", self._trigger_push_offxs
		)
		self._push_offys_binding = self.fbind(
			"offys", self._trigger_push_offys
		)

		def upd_box_translate(*_):
			self.box_translate.xy = self.pos

		def upd_box_points(*_):
			self.box.points = [
				0,
				0,
				self.width,
				0,
				self.width,
				self.height,
				0,
				self.height,
				0,
				0,
			]

		self.boxgrp = boxgrp = InstructionGroup()
		self.color = Color(*self.linecolor)
		self.box_translate = Translate(*self.pos)
		boxgrp.add(PushMatrix())
		boxgrp.add(self.box_translate)
		boxgrp.add(self.color)
		self.box = Line()
		upd_box_points()
		self._upd_box_points_binding = self.fbind("size", upd_box_points)
		self._upd_box_translate_binding = self.fbind("pos", upd_box_translate)
		boxgrp.add(self.box)
		boxgrp.add(Color(1.0, 1.0, 1.0))
		boxgrp.add(PopMatrix())
		self._finalized = True

	@logwrap(section="GraphPawnSpot")
	def unfinalize(self):
		self.unbind_uid("paths", self._push_image_paths_binding)
		self.unbind_uid("offxs", self._push_offxs_binding)
		self.unbind_uid("offys", self._push_offys_binding)
		self._finalized = False

	@logwrap(section="GraphPawnSpot")
	def pull_from_proxy(self, *_):
		initial = not hasattr(self, "_finalized")
		self.unfinalize()
		for key, att in [
			("_image_paths", "paths"),
			("_offxs", "offxs"),
			("_offys", "offys"),
		]:
			if key in self.proxy and self.proxy[key] != getattr(self, att):
				setattr(self, att, self.proxy[key])
		self.finalize(initial)

	@logwrap(section="GraphPawnSpot")
	def _trigger_pull_from_proxy(self, *_, **__):
		Clock.unschedule(self.pull_from_proxy)
		Clock.schedule_once(self.pull_from_proxy, 0)

	@trigger
	@logwrap(section="GraphPawnSpot")
	def _trigger_push_image_paths(self, *_):
		self.proxy["_image_paths"] = list(self.paths)

	@trigger
	@logwrap(section="GraphPawnSpot")
	def _trigger_push_offxs(self, *_):
		self.proxy["_offxs"] = list(self.offxs)

	@trigger
	@logwrap(section="GraphPawnSpot")
	def _trigger_push_offys(self, *_):
		self.proxy["_offys"] = list(self.offys)

	@logwrap(section="GraphPawnSpot")
	def on_linecolor(self, *_):
		"""If I don't yet have the instructions for drawing the selection box
		in my canvas, put them there. In any case, set the
		:class:`Color` instruction to match my current ``linecolor``.

		"""
		if hasattr(self, "color"):
			self.color.rgba = self.linecolor

	@logwrap(section="GraphPawnSpot")
	def on_board(self, *_):
		if not (hasattr(self, "group") and hasattr(self, "boxgrp")):
			Clock.schedule_once(self.on_board, 0)
			return
		self.canvas.add(self.group)
		self.canvas.add(self.boxgrp)

	@logwrap(section="GraphPawnSpot")
	def add_widget(self, wid, index=None, canvas=None):
		if index is None:
			for index, child in enumerate(self.children, start=1):
				if wid.priority < child.priority:
					index = len(self.children) - index
					break
		super().add_widget(wid, index=index, canvas=canvas)
		self._trigger_layout()

	@logwrap(section="GraphPawnSpot")
	def do_layout(self, *_):
		# First try to lay out my children inside of me,
		# leaving at least this much space on the sides
		xpad = self.proxy.get("_xpad", self.width / 4)
		ypad = self.proxy.get("_ypad", self.height / 4)
		self.gutter = gutter = self.proxy.get("_gutter", xpad / 2)
		height = self.height - ypad
		content_height = 0
		too_tall = False
		width = self.width - xpad
		content_width = 0
		groups = defaultdict(list)
		for child in self.children:
			group = child.proxy.get("_group", "")
			groups[group].append(child)
			if child.height > height:
				height = child.height
				too_tall = True
		piles = {}
		# Arrange the groups into piles that will fit in me vertically
		for group, members in groups.items():
			members.sort(key=lambda x: x.width * x.height, reverse=True)
			high = 0
			subgroups = []
			subgroup = []
			for member in members:
				high += member.height
				if high > height:
					subgroups.append(subgroup)
					subgroup = [member]
					high = member.height
				else:
					subgroup.append(member)
			subgroups.append(subgroup)
			content_height = max(
				(content_height, sum(wid.height for wid in subgroups[0]))
			)
			content_width += sum(
				max(wid.width for wid in subgrp) for subgrp in subgroups
			)
			piles[group] = subgroups
		self.content_width = content_width + gutter * (len(piles) - 1)
		too_wide = content_width > width
		# If I'm big enough to fit all this stuff, calculate an offset that will ensure
		# it's all centered. Otherwise just offset to my top-right so the user can still
		# reach me underneath all the pawns.
		if too_wide:
			offx = self.width
		else:
			offx = self.width / 2 - content_width / 2
		if too_tall:
			offy = self.height
		else:
			offy = self.height / 2 - content_height / 2
		for pile, subgroups in sorted(piles.items()):
			for subgroup in subgroups:
				subw = subh = 0
				for member in subgroup:
					rel_y = offy + subh
					member.rel_pos = (offx, rel_y)
					x, y = self.pos
					member.pos = x + offx, y + rel_y
					subw = max((subw, member.width))
					subh += member.height
				offx += subw
			offx += gutter

	@logwrap(section="GraphPawnSpot")
	def _position(self, *_):
		x, y = self.pos
		for child in self.children:
			offx, offy = getattr(child, "rel_pos", (0, 0))
			child.pos = x + offx, y + offy

	@logwrap(section="GraphPawnSpot")
	def on_selected(self, *_):
		if self.selected:
			self.linecolor = self.selected_outline_color
		else:
			self.linecolor = self.unselected_outline_color



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import mainthread
from kivy.graphics import Color, Line
from kivy.properties import NumericProperty, ObjectProperty, StringProperty
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.recycleview import RecycleView

from .util import logwrap, store_kv


class RuleStepper(RecycleView):
	name = StringProperty()

	@logwrap(section="RuleStepper")
	def from_rules_handled_turn(self, rules_handled_turn):
		data = [
			{
				"widget": "RuleStepperRuleButton",
				"name": "start",
				"end_tick": 0,
				"height": 40,
			}
		]
		all_rules = []
		for rbtyp, rules in rules_handled_turn.items():
			for tick, rule in rules.items():
				all_rules.append((tick, rbtyp, rule))
		all_rules.sort()
		prev_tick = 0
		last_entity = None
		last_rulebook = None
		lasttyp = None
		for tick, rbtyp, (entity, rulebook, rule) in all_rules:
			if tick == prev_tick:
				continue  # Rules that aren't triggered are still "handled". Ignore them.
			if lasttyp != rbtyp:
				data.append({"widget": "RulebookTypeLabel", "name": rbtyp})
			lasttyp = rbtyp
			rulebook_per_entity = rbtyp in {"thing", "place", "portal"}
			if not rulebook_per_entity:
				if rulebook != last_rulebook:
					last_rulebook = rulebook
					data.append({"widget": "RulebookLabel", "name": rulebook})
			if entity != last_entity:
				last_entity = entity
				data.append({"widget": "EntityLabel", "name": entity})
			if rulebook_per_entity:
				if rulebook != last_rulebook:
					rulebook = last_rulebook
					data.append({"widget": "RulebookLabel", "name": rulebook})
			data.append(
				{
					"widget": "RuleStepperRuleButton",
					"name": rule,
					"start_tick": prev_tick,
					"end_tick": tick,
					"height": 40,
				}
			)
			prev_tick = tick
		self.data = data


class RuleStepperRuleButton(Button):
	name = StringProperty()
	start_tick = NumericProperty()
	end_tick = NumericProperty()
	tick = NumericProperty()
	set_tick = ObjectProperty()

	def __init__(self, **kwargs):
		super(RuleStepperRuleButton, self).__init__(**kwargs)
		self.bind(pos=self.upd_line, size=self.upd_line, tick=self.upd_line)

	@logwrap(section="RuleStepperRuleButton")
	def on_release(self, *args):
		self.set_tick(self.end_tick)
		self.tick = self.end_tick

	@mainthread
	@logwrap(section="RuleStepperRuleButton")
	def upd_line(self, *_):
		if hasattr(self, "color_inst"):
			if self.tick == self.end_tick:
				self.color_inst.rgba = [1, 0, 0, 1]
				self.line.points = [self.x, self.y, self.right, self.y]
			else:
				self.color_inst.rgba = [0, 0, 0, 0]
		else:
			with self.canvas:
				self.color_inst = Color(
					rgba=(
						[1, 0, 0, 1]
						if self.tick in (self.start_tick, self.end_tick)
						else [0, 0, 0, 0]
					)
				)
				self.line = Line(
					points=[self.x, self.top, self.right, self.top]
				)


class EntityLabel(Label):
	name = ObjectProperty()


class RulebookLabel(Label):
	name = ObjectProperty()  # rulebooks may have tuples for names


class RulebookTypeLabel(Label):
	name = StringProperty()


store_kv(
	__name__,
	"""
#:import ScrollEffect kivy.effects.scroll.ScrollEffect
<RuleStepper>:
	key_viewclass: 'widget'
	effect_cls: ScrollEffect
	RecycleGridLayout:
		cols: 1
		size_hint_y: None
		default_size_hint: 1, None
		default_height: 20
		height: self.minimum_height
<RuleStepperRuleButton>:
	text: '\\n'.join((self.name, str(self.end_tick)))
	font_size: 14
	text_size: self.width, None
	halign: 'center'
	tick: app.tick
	set_tick: app.time_travel_to_tick
<EntityLabel>:
	multiline: True
	text: str(self.name)
	text_size: self.width, None
	size: self.texture_size
	font_size: 14
	padding_x: 8
<RulebookLabel>:
	text: str(self.name)
	text_size: self.width, None
	size: self.texture_size
	font_size: 14
	bold: True
	padding_x: 4
<RulebookTypeLabel>:
	text: self.name
	text_size: self.width, None
	font_size: 16
	bold: True
	size: self.texture_size
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget to represent :class:`Place`s. :class:`Pawn` moves around on
top of these.

"""

from kivy.clock import Clock, triggered

from elide.pawnspot import GraphPawnSpot

from .arrow import get_points, get_quad_vertices

try:
	from kivy_garden.collider import Collide2DPoly
except ImportError:
	from ..collide import Collide2DPoly


def trigger(func):
	return triggered()(func)


class GraphSpot(GraphPawnSpot):
	"""The icon that represents a :class:`Place`.

	Each :class:`Spot` is located on the Board that represents the
	:class:`Character` that the underlying :class:`Place` is in. Its
	coordinates are relative to its :class:`Board`, not necessarily
	the window the :class:`Board` is in.

	"""

	default_image_paths = ["atlas://rltiles/floor.atlas/floor-stone"]
	default_pos = (0.5, 0.5)

	def __init__(self, **kwargs):
		"""Deal with triggers and bindings, and arrange to take care of
		changes in game-time.

		"""
		self._pospawn_partials = {}
		self._pospawn_triggers = {}
		kwargs["size_hint"] = (None, None)
		if "place" in kwargs:
			kwargs["proxy"] = kwargs["place"]
			del kwargs["place"]
		super().__init__(**kwargs)

	def on_board(self, *args):
		super().on_board(*args)
		self.board.bind(size=self._upd_pos)

	def on_pos(self, *args):
		def upd(orig, dest):
			bgr = r * bg_scale_selected  # change for selectedness pls
			if (orig, dest) not in port_index:
				return
			idx = port_index[orig, dest]
			inst = instructions[orig, dest]
			(ox, oy, dx, dy), (x1, y1, endx, endy, x2, y2) = get_points(
				spot[orig], spot[dest], arrowhead_size
			)
			if ox < dx:
				bot_left_xs[idx] = ox - bgr
				top_right_xs[idx] = dx + bgr
			else:
				bot_left_xs[idx] = dx - bgr
				top_right_xs[idx] = ox + bgr
			if oy < dy:
				bot_left_ys[idx] = oy - bgr
				top_right_ys[idx] = dy + bgr
			else:
				bot_left_ys[idx] = dy - bgr
				top_right_ys[idx] = oy + bgr
			quadverts = get_quad_vertices(
				ox, oy, dx, dy, x1, y1, endx, endy, x2, y2, bgr, r
			)
			inst["shaft_bg"].points = quadverts["shaft_bg"]
			colliders[orig, dest] = Collide2DPoly(points=quadverts["shaft_bg"])
			inst["left_head_bg"].points = quadverts["left_head_bg"]
			inst["right_head_bg"].points = quadverts["right_head_bg"]
			inst["shaft_fg"].points = quadverts["shaft_fg"]
			inst["left_head_fg"].points = quadverts["left_head_fg"]
			inst["right_head_fg"].points = quadverts["right_head_fg"]

		if not self.board:
			return
		arrow_plane = self.board.arrow_plane
		fbo = arrow_plane._fbo
		arrowhead_size = arrow_plane.arrowhead_size
		r = arrow_plane.arrow_width // 2
		bg_scale_selected = arrow_plane.bg_scale_selected
		spot = self.board.spot
		succ = self.board.arrow
		pred = self.board.pred_arrow
		name = self.name
		instructions = arrow_plane._instructions_map
		colliders = arrow_plane._colliders_map
		instructions = arrow_plane._instructions_map
		port_index = arrow_plane._port_index
		bot_left_xs = arrow_plane._bot_left_corner_xs
		bot_left_ys = arrow_plane._bot_left_corner_ys
		top_right_xs = arrow_plane._top_right_corner_xs
		top_right_ys = arrow_plane._top_right_corner_ys
		fbo.bind()
		fbo.clear_buffer()
		if name in succ:
			for dest in succ[name]:
				upd(name, dest)
		if name in pred:
			for orig in pred[name]:
				upd(orig, name)
		fbo.release()
		return super().on_pos(*args)

	def _upd_pos(self, *args):
		if self.board is None:
			Clock.schedule_once(self._upd_pos, 0)
			return
		self.pos = (
			int(self.proxy.get("_x", self.default_pos[0]) * self.board.width),
			int(self.proxy.get("_y", self.default_pos[1]) * self.board.height),
		)

	def finalize(self, initial=True):
		if initial:
			self._upd_pos()
		super().finalize(initial)

	def push_pos(self, *args):
		"""Set my current position, expressed as proportions of the graph's
		width and height, into the ``_x`` and ``_y`` keys of the
		entity in my ``proxy`` property, such that it will be
		recorded in the database.

		"""
		self.proxy["_x"] = self.x / self.board.width
		self.proxy["_y"] = self.y / self.board.height

	_trigger_push_pos = trigger(push_pos)

	def on_touch_up(self, touch):
		if touch.grab_current is not self:
			return False
		self.center = touch.pos
		self._trigger_push_pos()
		touch.ungrab(self)
		self._trigger_push_pos()
		return True

	def __repr__(self):
		"""Give my name and position."""
		return "<{}@({},{}) at {}>".format(self.name, self.x, self.y, id(self))



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget to display the contents of a :class:`kivy.atlas.Atlas` in
one :class:`kivy.uix.togglebutton.ToggleButton` apiece, arranged in a
:class:`kivy.uix.stacklayout.StackLayout`. The user selects graphics
from the :class:`Pallet`, and the :class:`Pallet` updates its
``selection`` list to show what the user selected."""

from kivy.atlas import Atlas
from kivy.clock import Clock, mainthread, triggered
from kivy.graphics import Rectangle
from kivy.logger import Logger
from kivy.properties import (
	DictProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	OptionProperty,
	ReferenceListProperty,
	StringProperty,
)
from kivy.resources import resource_find
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.stacklayout import StackLayout
from kivy.uix.togglebutton import ToggleButton

from .util import logwrap, store_kv


def trigger(func):
	return triggered()(func)


class SwatchButton(ToggleButton):
	"""Toggle button containing a texture and its name, which, when
	toggled, will report the fact to the :class:`Pallet` it's in.

	"""

	tex = ObjectProperty()
	"""Texture to display here"""

	@logwrap(section="SwatchButton")
	def on_state(self, *_):
		if self.state == "down":
			assert self not in self.parent.selection
			if self.parent.selection_mode == "single":
				for wid in self.parent.selection:
					if wid is not self:
						wid.state = "normal"
				self.parent.selection = [self]
			else:
				self.parent.selection.append(self)
		else:
			if self in self.parent.selection:
				self.parent.selection.remove(self)


class Pallet(StackLayout):
	"""Many :class:`SwatchButton`, gathered from an :class:`kivy.atlas.Atlas`."""

	atlas = ObjectProperty()
	""":class:`kivy.atlas.Atlas` object I'll make :class:`SwatchButton` from."""
	filename = StringProperty()
	"""Path to an atlas; will construct :class:`kivy.atlas.Atlas` when set"""
	swatches = DictProperty({})
	""":class:`SwatchButton` widgets here, keyed by name of their graphic"""
	swatch_width = NumericProperty(100)
	"""Width of each and every :class:`SwatchButton` here"""
	swatch_height = NumericProperty(75)
	"""Height of each and every :class:`SwatchButton` here"""
	swatch_size = ReferenceListProperty(swatch_width, swatch_height)
	"""Size of each and every :class:`SwatchButton` here"""
	selection = ListProperty([])
	"""List of :class:`SwatchButton`s that are selected"""
	selection_mode = OptionProperty("single", options=["single", "multiple"])
	"""Whether to allow only a 'single' selected :class:`SwatchButton` (default), or 'multiple'"""

	def __init__(self, **kwargs):
		self._trigger_upd_textures = Clock.create_trigger(self.upd_textures)
		super().__init__(**kwargs)

	def on_selection(self, *_):
		Logger.debug(
			"Pallet: {} got selection {}".format(self.filename, self.selection)
		)

	def on_filename(self, *_):
		if not self.filename:
			return
		resource = resource_find(self.filename)
		if not resource:
			raise ValueError("Couldn't find atlas: {}".format(self.filename))
		self.atlas = Atlas(resource)

	def on_atlas(self, *_):
		if self.atlas is None:
			return
		self.upd_textures()
		self.atlas.bind(textures=self._trigger_upd_textures)

	def upd_textures(self, *_):
		"""Create one :class:`SwatchButton` for each texture"""
		if self.canvas is None:
			Clock.schedule_once(self.upd_textures, 0)
			return
		swatches = self.swatches
		atlas_textures = self.atlas.textures
		remove_widget = self.remove_widget
		add_widget = self.add_widget
		swatch_size = self.swatch_size
		for name, swatch in list(swatches.items()):
			if name not in atlas_textures:
				remove_widget(swatch)
				del swatches[name]
		for name, tex in atlas_textures.items():
			if name in swatches and swatches[name] != tex:
				remove_widget(swatches[name])
			if name not in swatches or swatches[name] != tex:
				swatches[name] = SwatchButton(
					text=name,
					tex=tex,
					size_hint=(None, None),
					size=swatch_size,
				)
				add_widget(swatches[name])


store_kv(
	__name__,
	"""
<SwatchButton>:
	canvas:
		Rectangle:
			pos:
				(
				root.x + (root.width / 2 - root.tex.size[0] / 2) if root.tex else 0,
				root.y + root.height - root.tex.size[1] if root.tex else 0
				)
			size: root.tex.size
			texture: root.tex
<Pallet>:
	orientation: 'lr-tb'
	padding_y: 100
	size_hint: (None, None)
	height: self.minimum_height
""",
)


class PalletBox(BoxLayout):
	pallets = ListProperty()



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial, wraps

from kivy import Logger
from kivy.app import App
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.layout import LayoutSelectionBehavior

from lisien.util import repr_call_sig

loaded_kv = set()
KV = {}


def store_kv(name: str, kv: str) -> None:
	if name in KV:
		raise KeyError("Already have that kv", name)
	print("storing kv: " + name)
	KV[name] = kv


def load_kv(name: str) -> None:
	if name in loaded_kv:
		return
	if name not in KV:
		Logger.error("No kv: " + name)
		Logger.error("I've stored: " + ", ".join(KV.keys()))
	print("loading kv: " + name)
	Builder.load_string(KV[name])
	loaded_kv.add(name)


class SelectableRecycleBoxLayout(
	FocusBehavior, LayoutSelectionBehavior, RecycleBoxLayout
):
	pass


def dummynum(character, name):
	"""Count how many nodes there already are in the character whose name
	starts the same.

	"""
	num = 0
	for nodename in character.node:
		nodename = str(nodename)
		if nodename[: len(name)] != name:
			continue
		try:
			nodenum = int(nodename.lstrip(name))
		except ValueError:
			continue
		num = max((nodenum, num))
	return num


def logwrap(func=None, *, section="ElideApp"):
	if func is None:
		return partial(logwrap, section=section)

	@wraps(func)
	def fn(*args, **kwargs):
		Logger.debug(section + ": " + repr_call_sig(func, *args, **kwargs))
		try:
			ret = func(*args, **kwargs)
		finally:
			for handler in Logger.handlers:
				# ensure any files get sync'd
				if hasattr(handler, "fd"):
					handler.fd.flush()
		return ret

	return fn



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import sys

from elide.app import ElideApp


def elide():
	kwargs = {}
	if os.path.isdir(sys.argv[-1]):
		kwargs["prefix"] = sys.argv[-1]
		kwargs["immediate_start"] = True
	app = ElideApp(**kwargs)
	app.run()


if __name__ == "__main__":
	elide()



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The big layout that you view all of elide through.

Handles touch, selection, and time control. Contains a graph, a stat
grid, the time control panel, and the menu.

"""

from ast import literal_eval
from functools import partial
from threading import Thread

from kivy.app import App
from kivy.clock import Clock, mainthread, triggered
from kivy.factory import Factory
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.properties import (
	BooleanProperty,
	BoundedNumericProperty,
	DictProperty,
	ListProperty,
	NumericProperty,
	ObjectProperty,
	ReferenceListProperty,
	StringProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.screenmanager import Screen
from kivy.uix.scrollview import ScrollView
from kivy.uix.slider import Slider
from kivy.uix.togglebutton import ToggleButton

from .calendar import Agenda
from .charmenu import CharMenu
from .graph.board import GraphBoardView
from .grid.board import GridBoardView
from .stepper import RuleStepper
from .util import dummynum, logwrap, store_kv


def trigger(func):
	return triggered()(func)


Factory.register("CharMenu", cls=CharMenu)


def release_edit_lock(*_):
	app = App.get_running_app()
	app.edit_locked = app.simulate_button_down


class KvLayout(FloatLayout):
	pass


class StatListPanel(BoxLayout):
	"""A panel that displays a simple two-column grid showing the stats of
	the selected entity, defaulting to those of the character being
	viewed.

	Has a button on the bottom to open the StatWindow in which
	to add and delete stats, or to change the way they are displayed
	in the StatListPanel.

	"""

	selection_name = StringProperty()
	button_text = StringProperty("Configure stats")
	cfgstatbut = ObjectProperty()
	statlist = ObjectProperty()
	proxy = ObjectProperty()
	toggle_stat_cfg = ObjectProperty()

	@logwrap(section="StatListPanel")
	def on_proxy(self, *_):
		if hasattr(self.proxy, "name"):
			self.selection_name = str(self.proxy.name)

	@logwrap(section="StatListPanel")
	def set_value(self, k, v):
		if v is None:
			del self.proxy[k]
		else:
			try:
				vv = literal_eval(v)
			except (TypeError, ValueError):
				vv = v
			self.proxy[k] = vv


class SimulateButton(ToggleButton):
	@logwrap(section="SimulateButton")
	def on_state(self, *_):
		app = App.get_running_app()
		app.edit_locked = app.simulate_button_down = self.state == "down"


class OneTurnButton(Button):
	screen = ObjectProperty()

	@logwrap(section="OneTurnButton")
	def on_release(self):
		App.get_running_app().edit_locked = True
		self.screen.next_turn(cb=release_edit_lock)


class TimePanel(BoxLayout):
	"""A panel that starts and stop the game, or sets the time.

	There's a "simulate" button, which is toggleable. When toggled on, the
	simulation will continue to run until it's toggled off
	again. Next to this is a "1 turn" button, which will simulate
	exactly one turn and stop. And there are two text fields in which
	you can manually enter a Branch and Tick to go to. Moving through
	time this way doesn't simulate anything--you'll only see what
	happened as a result of "simulate," "1 turn," or some other way
	the lisien rules engine has been made to run.

	"""

	screen = ObjectProperty()
	buttons_font_size = NumericProperty(18)
	disable_one_turn = BooleanProperty()

	@logwrap(section="TimePanel")
	def set_branch(self, *_):
		branch = self.ids.branchfield.text
		self.ids.branchfield.text = ""
		self.screen.app.branch = branch
		self.screen.charmenu.switch_to_menu()

	@logwrap(section="TimePanel")
	def set_turn(self, *_):
		turn = int(self.ids.turnfield.text)
		self.ids.turnfield.text = ""
		self.screen.app.set_turn(turn)
		self.screen.charmenu.switch_to_menu()

	@logwrap(section="TimePanel")
	def set_tick(self, *_):
		tick = int(self.ids.tickfield.text)
		self.ids.tickfield.text = ""
		self.screen.app.set_tick(tick)
		self.screen.charmenu.switch_to_menu()

	@mainthread
	@logwrap(section="TimePanel")
	def _upd_branch_hint(self, app, *_):
		self.ids.branchfield.hint_text = app.branch

	@mainthread
	@logwrap(section="TimePanel")
	def _upd_turn_hint(self, app, *_):
		self.ids.turnfield.hint_text = str(app.turn)

	@mainthread
	@logwrap(section="TimePanel")
	def _upd_tick_hint(self, app, *_):
		self.ids.tickfield.hint_text = str(app.tick)

	@logwrap(section="TimePanel")
	def on_screen(self, *_):
		if not all(
			field in self.ids
			for field in ("branchfield", "turnfield", "tickfield")
		):
			Clock.schedule_once(self.on_screen, 0)
			return
		self.ids.branchfield.hint_text = self.screen.app.branch
		self.ids.turnfield.hint_text = str(self.screen.app.turn)
		self.ids.tickfield.hint_text = str(self.screen.app.tick)
		self.screen.app.bind(
			branch=self._upd_branch_hint,
			turn=self._upd_turn_hint,
			tick=self._upd_tick_hint,
		)


class MainScreen(Screen):
	"""A master layout that contains one graph and some menus.

	This contains three elements: a scrollview (containing the graph),
	a menu, and the time control panel. This class has some support methods
	for handling interactions with the menu and the character sheet,
	but if neither of those happen, the scrollview handles touches on its
	own.

	"""

	name = "mainscreen"

	graphboards = DictProperty()
	gridboards = DictProperty()
	boardview = ObjectProperty()
	mainview = ObjectProperty()
	charmenu = ObjectProperty()
	statlist = ObjectProperty()
	statpanel = ObjectProperty()
	timepanel = ObjectProperty()
	turnscroll = ObjectProperty()
	kv = StringProperty()
	use_kv = BooleanProperty()
	play_speed = NumericProperty(1)
	playbut = ObjectProperty()
	portaladdbut = ObjectProperty()
	portaldirbut = ObjectProperty()
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	dummies = ReferenceListProperty(dummyplace, dummything)
	dialoglayout = ObjectProperty()
	visible = BooleanProperty()
	_touch = ObjectProperty(None, allownone=True)
	rules_per_frame = BoundedNumericProperty(10, min=1)
	tmp_block = BooleanProperty(False)

	@property
	def app(self):
		return App.get_running_app()

	@logwrap(section="TimePanel")
	def _update_adding_portal(self, *_):
		self.boardview.adding_portal = (
			self.charmenu.portaladdbut.state == "down"
		)

	@logwrap(section="TimePanel")
	def _update_board(self, *_):
		self.boardview.board = self.graphboards[self.app.character_name]
		self.gridview.board = self.gridboards[self.app.character_name]

	@logwrap(section="TimePanel")
	def populate(self, *_):
		if hasattr(self, "_populated"):
			return
		if (
			None in (self.statpanel, self.charmenu)
			or None in (self.app.character_name, self.charmenu.portaladdbut)
			or self.app.character_name not in self.graphboards
		):
			if self.statpanel is None:
				Logger.warning("MainScreen: no statpanel")
			if self.charmenu is None:
				Logger.warning("MainScreen: no charmenu")
			elif self.charmenu.portaladdbut is None:
				Logger.warning("MainScreen: CharMenu has no portal button")
			if self.app.character_name is None:
				Logger.warning("MainScreen: no character chosen")
			elif self.app.character_name not in self.graphboards:
				Logger.warning(
					"MainScreen: no graphboard to represent "
					+ repr(self.app.character_name)
				)
			Clock.schedule_once(self.populate, 0)
			return
		self.boardview = GraphBoardView(
			scale_min=0.2,
			scale_max=4.0,
			size=self.mainview.size,
			pos=self.mainview.pos,
			board=self.graphboards[self.app.character_name],
			adding_portal=self.charmenu.portaladdbut.state == "down",
		)
		self.mainview.bind(
			size=self.boardview.setter("size"),
			pos=self.boardview.setter("pos"),
		)
		self.charmenu.portaladdbut.bind(state=self._update_adding_portal)
		self.app.bind(character_name=self._update_board)
		self.calendar = Agenda(update_mode="present")
		self.calendar_view = ScrollView(
			size=self.mainview.size, pos=self.mainview.pos
		)
		self.gridview = GridBoardView(
			scale_min=0.2,
			scale_max=4.0,
			size=self.mainview.size,
			pos=self.mainview.pos,
			board=self.gridboards[self.app.character_name],
		)
		self.mainview.bind(
			size=self.calendar_view.setter("size"),
			pos=self.calendar_view.setter("pos"),
		)
		self.mainview.bind(
			size=self.gridview.setter("size"), pos=self.gridview.setter("pos")
		)
		self.calendar_view.add_widget(self.calendar)
		self.mainview.add_widget(self.boardview)
		self._populated = True

	@logwrap(section="TimePanel")
	def on_statpanel(self, *_):
		if not self.app:
			Clock.schedule_once(self.on_statpanel, 0)
			return
		self.app.bind(
			selected_proxy=self._update_statlist,
			branch=self._update_statlist,
			turn=self._update_statlist,
			tick=self._update_statlist,
		)

	@trigger
	@logwrap(section="TimePanel")
	def _update_statlist(self, *_):
		if not self.app or not hasattr(self.app, "engine"):
			return
		if not self.app.selected_proxy:
			self._update_statlist()
			return
		self.app.update_calendar(
			self.statpanel.statlist, past_turns=0, future_turns=0
		)

	@logwrap(section="TimePanel")
	def pull_visibility(self, *_):
		if not self.manager:
			self.visible = False
			return
		self.visible = self.manager.current == "main"

	@logwrap(section="TimePanel")
	def on_manager(self, *_):
		if not self.manager:
			return
		self.pull_visibility()
		self.manager.bind(current=self.pull_visibility)

	@logwrap(section="TimePanel")
	def on_playbut(self, *_):
		if hasattr(self, "_play_scheduled"):
			return
		self._play_scheduled = Clock.schedule_interval(
			self.play, 1.0 / self.play_speed
		)

	@logwrap(section="TimePanel")
	def on_play_speed(self, *_):
		"""Change the interval at which ``self.play`` is called to match my
		current ``play_speed``.

		"""
		if hasattr(self, "_play_scheduled"):
			Clock.unschedule(self._play_scheduled)
		self._play_scheduled = Clock.schedule_interval(
			self.play, 1.0 / self.play_speed
		)

	@logwrap(section="TimePanel")
	def remake_display(self, *_):
		"""Remake any affected widgets after a change in my ``kv``."""
		Builder.load_string(self.kv)
		if hasattr(self, "_kv_layout"):
			self.remove_widget(self._kv_layout)
			del self._kv_layout
		self._kv_layout = KvLayout()
		self.add_widget(self._kv_layout)

	_trigger_remake_display = trigger(remake_display)

	@logwrap(section="TimePanel")
	def on_touch_down(self, touch):
		if self.visible:
			touch.grab(self)
		for interceptor in (
			self.timepanel,
			self.turnscroll,
			self.charmenu,
			self.statpanel,
			self.dummyplace,
			self.dummything,
		):
			if interceptor.collide_point(*touch.pos):
				interceptor.dispatch("on_touch_down", touch)
				self.boardview.keep_selection = (
					self.gridview.keep_selection
				) = True
				return True
		if self.dialoglayout.dispatch("on_touch_down", touch):
			return True
		return self.mainview.dispatch("on_touch_down", touch)

	@logwrap(section="TimePanel")
	def on_touch_up(self, touch):
		if self.timepanel.collide_point(*touch.pos):
			return self.timepanel.dispatch("on_touch_up", touch)
		elif self.turnscroll.collide_point(*touch.pos):
			return self.turnscroll.dispatch("on_touch_up", touch)
		elif self.charmenu.collide_point(*touch.pos):
			return self.charmenu.dispatch("on_touch_up", touch)
		elif self.statpanel.collide_point(*touch.pos):
			return self.statpanel.dispatch("on_touch_up", touch)
		return self.mainview.dispatch("on_touch_up", touch)

	@logwrap(section="TimePanel")
	def on_dummies(self, *_):
		"""Give the dummies numbers such that, when appended to their names,
		they give a unique name for the resulting new
		:class:`graph.Pawn` or :class:`graph.Spot`.

		"""
		if not self.app.character:
			Clock.schedule_once(self.on_dummies, 0)
			return

		def renum_dummy(dummy, *_):
			dummy.num = dummynum(self.app.character, dummy.prefix) + 1

		for dummy in self.dummies:
			if dummy is None or hasattr(dummy, "_numbered"):
				continue
			if dummy == self.dummything:
				self.app.pawncfg.bind(imgpaths=dummy.setter("paths"))
			if dummy == self.dummyplace:
				self.app.spotcfg.bind(imgpaths=dummy.setter("paths"))
			dummy.num = dummynum(self.app.character, dummy.prefix) + 1
			Logger.debug("MainScreen: dummy #{}".format(dummy.num))
			dummy.bind(prefix=partial(renum_dummy, dummy))
			dummy._numbered = True

	@logwrap(section="TimePanel")
	def update_from_time_travel(
		self, command, branch, turn, tick, result, **kwargs
	):
		self._update_from_delta(command, branch, turn, tick, result[-1])

	@logwrap(section="TimePanel")
	def _update_from_delta(self, cmd, branch, turn, tick, delta, **kwargs):
		self.app.branch = branch
		self.app.turn = turn
		self.app.tick = tick
		self.statpanel.statlist.mirror = dict(self.app.selected_proxy)

	@logwrap(section="TimePanel")
	def play(self, *_):
		"""If the 'play' button is pressed, advance a turn.

		If you want to disable this, set ``engine.universal['block'] = True``

		"""
		if (
			self.playbut is None
			or self.playbut.state == "normal"
			or not hasattr(self.app, "engine")
			or self.app.engine is None
			or self.app.engine.closed
			or self.app.engine.universal.get("block")
			or not hasattr(self.app, "manager")
			or self.app.manager.current != "mainscreen"
		):
			return
		self.next_turn(cb=release_edit_lock)

	@logwrap(section="TimePanel")
	def _update_from_next_turn(
		self, command, branch, turn, tick, result, cb=None
	):
		todo, deltas = result
		if isinstance(todo, list):
			self.dialoglayout.todo = todo
			self.dialoglayout.idx = 0
		self._update_from_delta(command, branch, turn, tick, deltas)
		self.dialoglayout.advance_dialog()
		if cb is not None:
			cb(command, branch, turn, tick, result)
		self.tmp_block = False

	@logwrap(section="TimePanel")
	def next_turn(self, cb=None, *_):
		"""Advance time by one turn, if it's not blocked.

		Block time by setting ``engine.universal['block'] = True``"""
		if self.tmp_block:
			return
		eng = self.app.engine
		dial = self.dialoglayout
		if eng.universal.get("block"):
			Logger.info(
				"MainScreen: next_turn blocked, delete universal['block'] to unblock"
			)
			return
		if dial.todo:
			if not dial.children:
				Logger.info(
					"MainScreen: DialogLayout has todo but no children, advancing."
				)
				dial.advance_dialog()
			Clock.schedule_once(partial(self.next_turn, cb), self.play_speed)
			return
		self.tmp_block = True
		self._next_turn_thread = Thread(
			target=eng.next_turn,
			kwargs={"cb": partial(self._update_from_next_turn, cb=cb)},
		)
		self._next_turn_thread.start()
		self.ids.charmenu.switch_to_menu()

	@logwrap(section="TimePanel")
	def switch_to_calendar(self, *_):
		self.app.update_calendar(self.calendar)
		self.mainview.clear_widgets()
		self.mainview.add_widget(self.calendar_view)

	@logwrap(section="TimePanel")
	def switch_to_boardview(self, *_):
		self.mainview.clear_widgets()
		self.app.engine.handle(
			"apply_choices", choices=[self.calendar.get_track()]
		)
		self.mainview.add_widget(self.boardview)

	@logwrap(section="TimePanel")
	def toggle_gridview(self, *_):
		if self.gridview in self.mainview.children:
			self.mainview.clear_widgets()
			self.mainview.add_widget(self.boardview)
		else:
			self.mainview.clear_widgets()
			self.mainview.add_widget(self.gridview)

	@logwrap(section="TimePanel")
	def toggle_calendar(self, *_):
		# TODO decide how to handle switching between >2 view types
		if self.boardview in self.mainview.children:
			self.switch_to_calendar()
		else:
			self.switch_to_boardview()

	@trigger
	@logwrap(section="TimePanel")
	def toggle_timestream(self, *_):
		if self.manager.current != "timestream":
			self.manager.current = "timestream"
		else:
			self.manager.current = "mainscreen"


class CharMenuContainer(BoxLayout):
	screen = ObjectProperty()
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	portaladdbut = ObjectProperty()
	toggle_gridview = ObjectProperty()
	toggle_timestream = ObjectProperty()

	def __init__(self, **kwargs):
		super(CharMenuContainer, self).__init__(**kwargs)
		self.charmenu = CharMenu(screen=self.screen, size_hint_y=0.9)
		self.bind(screen=self.charmenu.setter("screen"))
		self.dummyplace = self.charmenu.dummyplace
		self.charmenu.bind(dummyplace=self.setter("dummyplace"))
		self.dummything = self.charmenu.dummything
		self.charmenu.bind(dummything=self.setter("dummything"))
		self.portaladdbut = self.charmenu.portaladdbut
		self.charmenu.bind(portaladdbut=self.setter("portaladdbut"))
		if self.toggle_gridview:
			self.charmenu = self.toggle_gridview
		self.bind(
			toggle_gridview=self.charmenu.setter("toggle_gridview"),
			toggle_timestream=self.charmenu.setter("toggle_timestream"),
		)
		self.stepper = RuleStepper(size_hint_y=0.9)
		self.button = Button(
			on_release=self._toggle_stepper,
			text="Rule\nstepper",
			size_hint_y=0.1,
		)
		app = App.get_running_app()
		app.bind(
			branch=self.switch_to_menu,
			turn=self.switch_to_menu,
			edit_locked=self.button.setter("disabled"),
		)

	@logwrap(section="TimePanel")
	def on_parent(self, *_):
		if (
			not self.screen
			or not hasattr(self, "charmenu")
			or not hasattr(self, "stepper")
			or not hasattr(self, "button")
		):
			Clock.schedule_once(self.on_parent, 0)
			return
		self.add_widget(self.charmenu)
		self.add_widget(self.button)

	@trigger
	@logwrap(section="TimePanel")
	def _toggle_stepper(self, *_):
		if self.charmenu in self.children:
			engine = self.screen.app.engine
			self.clear_widgets()
			self.stepper.from_rules_handled_turn(
				engine.handle("rules_handled_turn")
			)
			self.add_widget(self.stepper)
			self.button.text = "Menu"
		else:
			self.clear_widgets()
			self.add_widget(self.charmenu)
			self.button.text = "Rule stepper"
		self.add_widget(self.button)

	@trigger
	@logwrap(section="TimePanel")
	def switch_to_menu(self, *args):
		if self.charmenu not in self.children:
			self.clear_widgets()
			self.add_widget(self.charmenu)
			self.button.text = "Rule stepper"
			self.add_widget(self.button)


class TurnScroll(Slider):
	def __init__(self, **kwargs):
		kwargs["step"] = 1
		super().__init__(**kwargs)
		self._collect_engine()

	@logwrap(section="TimePanel")
	def _collect_engine(self, *args):
		app = App.get_running_app()
		if not hasattr(app, "engine"):
			Clock.schedule_once(self._collect_engine, 0)
			return
		engine = app.engine
		self.min = engine.branch_start_turn()
		self.max = engine.branch_end_turn()
		self.value = engine.turn
		engine.time.connect(self._receive_time)

	@logwrap(section="TimePanel")
	def _receive_time(self, engine, then, now):
		(_, turn, _) = now
		self.value = turn
		try:
			self.min = engine.branch_start_turn()
		except KeyError:
			self.min = turn
		try:
			self.max = engine.branch_end_turn()
		except KeyError:
			self.max = turn

	@logwrap(section="TimePanel")
	def on_touch_move(self, touch):
		if touch.grab_current == self:
			app = App.get_running_app()
			app.mainscreen.timepanel.ids.turnfield.hint_text = str(
				int(self.value)
			)
		return super().on_touch_move(touch)

	@logwrap(section="TimePanel")
	def on_touch_up(self, touch):
		if touch.grab_current == self:
			app = App.get_running_app()
			app.engine.time.disconnect(self._receive_time)
			app.time_travel(app.branch, int(self.value))
			app.engine.time.connect(self._receive_time)


store_kv(
	__name__,
	"""
#: import resource_find kivy.resources.resource_find
<StatListPanel>:
	orientation: 'vertical'
	statlist: statlist
	id: statpanel
	proxy: app.selected_proxy
	Label:
		size_hint_y: 0.05
		text: root.selection_name
		bold: True
	Calendar:
		id: statlist
		do_scroll_x: False
		do_scroll_y: True
		entity: root.proxy
		update_mode: 'present'
		disabled: app.edit_locked
	Button:
		id: cfgstatbut
		size_hint_y: 0.05
		text: root.button_text
		disabled: app.edit_locked
		on_release: root.toggle_stat_cfg()
<SimulateButton>:
	Image:
		x: root.center_x - root.width / 3
		y: root.y + root.font_size + 3
		width: root.width / 1.5
		height: root.height / 1.5
		source: 'right.png'
	Label:
		id: playlabel
		font_size: root.font_size
		center_x: root.center_x
		y: root.y
		size: self.texture_size
		text: 'Simulate'
<OneTurnButton>:
	Image:
		x: root.center_x - (root.width / 3)
		y: root.y + root.font_size + 3
		width: root.width / 1.5
		height: root.height / 1.5
		source: 'next.png' 
	Label:
		font_size: root.font_size
		center_x: root.center_x
		y: root.y
		size: self.texture_size
		text: '1 turn'
<TimePanel>:
	orientation: 'vertical'
	playbut: playbut
	BoxLayout:
		size_hint_y: 0.4
		BoxLayout:
			orientation: 'vertical'
			Label:
				size_hint_y: 0.4
				text: 'Branch'
			MenuTextInput:
				id: branchfield
				set_value: root.set_branch
				hint_text: root.screen.app.branch if root.screen else ''
				disabled: app.edit_locked
		BoxLayout:
			BoxLayout:
				orientation: 'vertical'
				Label:
					size_hint_y: 0.4
					text: 'Turn'
				MenuIntInput:
					id: turnfield
					set_value: root.set_turn
					hint_text: str(root.screen.app.turn) if root.screen else ''
					disabled: app.edit_locked
			BoxLayout:
				orientation: 'vertical'
				Label:
					size_hint_y: 0.4
					text: 'Tick'
				MenuIntInput:
					id: tickfield
					set_value: root.set_tick
					hint_text: str(root.screen.app.tick) if root.screen else ''
					disabled: app.edit_locked
	BoxLayout:
		size_hint_y: 0.6
		SimulateButton:
			id: playbut
			font_size: root.buttons_font_size
			disabled: self.state != 'down' and app.edit_locked
		OneTurnButton:
			id: stepbut
			font_size: root.buttons_font_size
			screen: root.screen
			disabled: root.disable_one_turn or app.edit_locked
<MainScreen>:
	name: 'mainscreen'
	dummyplace: charmenu.dummyplace
	dummything: charmenu.dummything
	mainview: mainview
	playbut: timepanel.playbut
	portaladdbut: charmenu.portaladdbut
	charmenu: charmenu
	statlist: statpanel.statlist
	statpanel: statpanel
	timepanel: timepanel
	turnscroll: turnscroll
	dialoglayout: dialoglayout
	TurnScroll:
		id: turnscroll
		value_track: True
		pos_hint: {'bot': 0}
		size_hint: (1, 0.1)
		disabled: app.edit_locked
	Widget:
		id: mainview
		x: statpanel.right
		y: turnscroll.top
		size_hint: (None, None)
		width: charmenu.x - statpanel.right
		height: root.height
	StatListPanel:
		id: statpanel
		toggle_stat_cfg: app.statcfg.toggle
		x: 0
		y: timepanel.top
		size_hint: (0.25, 0.7)
		selection_name: app.selected_proxy_name
		toggle_timestream: root.toggle_timestream
	TimePanel:
		id: timepanel
		screen: root
		x: 0
		y: turnscroll.top
		size_hint: (0.25, 0.2)
		disable_one_turn: root.tmp_block
	CharMenuContainer:
		id: charmenu
		toggle_calendar: root.toggle_calendar
		toggle_gridview: root.toggle_gridview
		toggle_timestream: root.toggle_timestream
		orientation: 'vertical'
		screen: root
		pos_hint: {'right': 1, 'top': 1}
		size_hint: (0.2, 0.9)
	DialogLayout:
		id: dialoglayout
		size_hint: None, None
		x: statpanel.right
		y: timepanel.top
		width: charmenu.x - statpanel.right
		height: root.height - timepanel.top
""",
)



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Grid of current values for some entity. Can be changed by the
user. Autoupdates when there's a change for any reason.

"""

from functools import partial

from kivy.clock import Clock
from kivy.properties import DictProperty, ObjectProperty
from kivy.uix.recycleview import RecycleView

from .util import logwrap

default_cfg = {
	"control": "readout",
	"true_text": "1",
	"false_text": "0",
	"min": 0.0,
	"max": 1.0,
}


class BaseStatListView(RecycleView):
	"""Base class for widgets showing lists of stats and their values"""

	proxy = ObjectProperty()
	"""A proxy object representing a lisien entity"""
	engine = ObjectProperty()
	"""A :class:`lisien.proxy.EngineProxy` object"""
	app = ObjectProperty()
	"""The Kivy app object"""
	_scheduled_set_value = DictProperty()

	def __init__(self, **kwargs):
		self._listeners = {}
		super().__init__(**kwargs)

	@logwrap(section="BaseStatListView")
	def on_proxy(self, *_):
		self.proxy.connect(self._trigger_upd_data, weak=False)
		self._trigger_upd_data()

	@logwrap(section="BaseStatListView")
	def del_key(self, k):
		"""Delete the key and any configuration for it"""
		if k not in self.proxy:
			raise KeyError
		del self.proxy[k]
		if "_config" in self.proxy and k in self.proxy["_config"]:
			del self.proxy["_config"][k]

	@logwrap(section="BaseStatListView")
	def set_value(self, k, v):
		"""Set a value on the proxy, parsing it to a useful datatype if possible"""
		from ast import literal_eval

		if self.engine is None or self.proxy is None:
			self._trigger_set_value(k, v)
			return
		if v is None:
			del self.proxy[k]
		else:
			try:
				vv = literal_eval(v)
			except (TypeError, ValueError):
				vv = v
			self.proxy[k] = vv
		if (k, v) in self._scheduled_set_value:
			del self._scheduled_set_value[k, v]

	@logwrap(section="BaseStatListView")
	def _trigger_set_value(self, k, v, *_):
		todo = partial(self.set_value, k, v)
		if (k, v) in self._scheduled_set_value:
			Clock.unschedule(self._scheduled_set_value[k, v])
		self._scheduled_set_value[k, v] = Clock.schedule_once(todo, 0)

	@logwrap(section="BaseStatListView")
	def init_config(self, key):
		"""Set the configuration for the key to something that will always work"""
		self.proxy["_config"].setdefault(key, default_cfg)

	@logwrap(section="BaseStatListView")
	def set_config(self, key, option, value):
		"""Set a configuration option for a key"""
		if "_config" not in self.proxy:
			newopt = dict(default_cfg)
			newopt[option] = value
			self.proxy["_config"] = {key: newopt}
		else:
			if key in self.proxy["_config"]:
				self.proxy["_config"][key][option] = value
			else:
				newopt = dict(default_cfg)
				newopt[option] = value
				self.proxy["_config"][key] = newopt

	@logwrap(section="BaseStatListView")
	def set_configs(self, key, d):
		"""Set the whole configuration for a key"""
		if "_config" in self.proxy:
			self.proxy["_config"][key] = d
		else:
			self.proxy["_config"] = {key: d}

	@logwrap(section="BaseStatListView")
	def iter_data(self):
		"""Iterate over key-value pairs that are really meant to be displayed"""
		invalid = {"character", "name", "location", "rulebooks"}
		for k, v in self.proxy.items():
			if not (isinstance(k, str) and k[0] == "_") and k not in invalid:
				yield k, v

	@logwrap(section="BaseStatListView")
	def munge(self, k, v):
		"""Turn a key and value into a dictionary describing a widget to show"""
		if "_config" in self.proxy and k in self.proxy["_config"]:
			config = self.proxy["_config"][k].unwrap()
		else:
			config = default_cfg
		return {
			"key": k,
			"reg": self._reg_widget,
			"unreg": self._unreg_widget,
			"gett": self.proxy.__getitem__,
			"sett": self.set_value,
			"listen": self.proxy.connect,
			"unlisten": self.proxy.disconnect,
			"config": config,
		}

	@logwrap(section="BaseStatListView")
	def upd_data(self, *_):
		"""Update to match new entity data"""
		data = [self.munge(k, v) for k, v in self.iter_data()]
		self.data = sorted(data, key=lambda d: d["key"])

	@logwrap(section="BaseStatListView")
	def _trigger_upd_data(self, *_, **__):
		if hasattr(self, "_scheduled_upd_data"):
			Clock.unschedule(self._scheduled_upd_data)
		self._scheduled_upd_data = Clock.schedule_once(self.upd_data, 0)

	@logwrap(section="BaseStatListView")
	def _reg_widget(self, w, *_):
		if not self.proxy:
			Clock.schedule_once(partial(self._reg_widget, w), 0)
			return

		def listen(*_):
			if w.key not in self.proxy:
				return
			if w.value != self.proxy[w.key]:
				w.value = self.proxy[w.key]

		self._listeners[w.key] = listen
		self.proxy.connect(listen)

	@logwrap(section="BaseStatListView")
	def _unreg_widget(self, w):
		if w.key in self._listeners:
			self.proxy.disconnect(self._listeners[w.key])



# This file is part of the kivy-garden project.
# Copyright (c) Zachary Spector, public@zacharyspector.com
# Available under the terms of the MIT license.
"""Several textures superimposed on one another, and possibly offset
by some amount.

In 2D games where characters can wear different clothes or hold
different equipment, their graphics are often composed of several
graphics layered on one another. This widget simplifies the management
of such compositions.

"""

from operator import itemgetter
from time import monotonic

import numpy as np
from kivy import Logger
from kivy.clock import Clock, mainthread
from kivy.core.image import Image
from kivy.graphics import (
	Color,
	Fbo,
	InstructionGroup,
	Line,
	PopMatrix,
	PushMatrix,
	Rectangle,
	Translate,
)
from kivy.graphics.fbo import Fbo
from kivy.properties import (
	AliasProperty,
	DictProperty,
	ListProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.resources import resource_find
from kivy.uix.widget import Widget


class TextureStack(Widget):
	"""Several textures superimposed on one another, and possibly offset
	by some amount.

	In 2D games where characters can wear different clothes or hold
	different equipment, their graphics are often composed of several
	graphics layered on one another. This widget simplifies the
	management of such compositions.

	"""

	texs = ListProperty()
	"""Texture objects"""
	offxs = ListProperty()
	"""x-offsets. The texture at the same index will be moved to the right
    by the number of pixels in this list.

    """
	offys = ListProperty()
	"""y-offsets. The texture at the same index will be moved upward by
    the number of pixels in this list.

    """
	group = ObjectProperty()
	"""My ``InstructionGroup``, suitable for addition to whatever ``canvas``."""

	def _get_offsets(self):
		return zip(self.offxs, self.offys)

	def _set_offsets(self, offs):
		offxs = []
		offys = []
		for x, y in offs:
			offxs.append(x)
			offys.append(y)
		self.offxs, self.offys = offxs, offys

	offsets = AliasProperty(
		_get_offsets, _set_offsets, bind=("offxs", "offys")
	)
	"""List of (x, y) tuples by which to offset the corresponding texture."""
	_texture_rectangles = DictProperty({})
	"""Private.

    Rectangle instructions for each of the textures, keyed by the
    texture.

    """

	def __init__(self, **kwargs):
		"""Make triggers and bind."""
		kwargs["size_hint"] = (None, None)
		self.translate = Translate(0, 0)
		self.group = InstructionGroup()
		super().__init__(**kwargs)
		self.bind(offxs=self.on_pos, offys=self.on_pos)

	def on_texs(self, *args):
		"""Make rectangles for each of the textures and add them to the canvas."""
		if not self.canvas or not self.texs:
			Clock.schedule_once(self.on_texs, 0)
			return
		texlen = len(self.texs)
		# Ensure each property is the same length as my texs, padding
		# with 0 as needed
		for prop in ("offxs", "offys"):
			proplen = len(getattr(self, prop))
			if proplen > texlen:
				setattr(self, prop, getattr(self, prop)[: proplen - texlen])
			if texlen > proplen:
				propval = list(getattr(self, prop))
				propval += [0] * (texlen - proplen)
				setattr(self, prop, propval)
		self.canvas.remove(self.group)
		self.group = InstructionGroup()
		self._texture_rectangles = {}
		w = h = 0
		(x, y) = self.pos
		self.translate.x = x
		self.translate.y = y
		self.group.add(PushMatrix())
		self.group.add(self.translate)
		for tex, offx, offy in zip(self.texs, self.offxs, self.offys):
			rect = Rectangle(pos=(offx, offy), size=tex.size, texture=tex)
			self._texture_rectangles[tex] = rect
			self.group.add(rect)
			tw = tex.width + offx
			th = tex.height + offy
			if tw > w:
				w = tw
			if th > h:
				h = th
		self.size = (w, h)
		self.group.add(PopMatrix())
		self.canvas.add(self.group)
		self.canvas.ask_update()

	def on_pos(self, *args):
		"""Translate all the rectangles within this widget to reflect the widget's position."""
		(x, y) = self.pos
		self.translate.x = x
		self.translate.y = y

	def clear(self):
		"""Clear my rectangles and ``texs``."""
		self.group.clear()
		self._texture_rectangles = {}
		self.texs = []
		self.size = [1, 1]

	def insert(self, i, tex):
		"""Insert the texture into my ``texs``, waiting for the creation of
		the canvas if necessary.

		"""
		if not self.canvas:
			Clock.schedule_once(lambda dt: self.insert(i, tex), 0)
			return
		self.texs.insert(i, tex)

	def append(self, tex):
		"""``self.insert(len(self.texs), tex)``"""
		self.insert(len(self.texs), tex)

	def __delitem__(self, i):
		"""Remove a texture and its rectangle"""
		tex = self.texs[i]
		try:
			rect = self._texture_rectangles[tex]
			self.canvas.remove(rect)
			del self._texture_rectangles[tex]
		except KeyError:
			pass
		del self.texs[i]

	def __setitem__(self, i, v):
		"""First delete at ``i``, then insert there"""
		if len(self.texs) > 0:
			self._no_upd_texs = True
			self.__delitem__(i)
			self._no_upd_texs = False
		self.insert(i, v)

	def pop(self, i=-1):
		"""Delete the texture at ``i``, and return it."""
		return self.texs.pop(i)


class ImageStack(TextureStack):
	"""Instead of supplying textures themselves, supply paths to where the
	textures may be loaded from.

	"""

	paths = ListProperty()
	"""List of paths to images you want stacked."""
	pathtexs = DictProperty()
	"""Private. Dictionary mapping image paths to textures of the images."""
	pathimgs = DictProperty()
	"""Dictionary mapping image paths to ``kivy.core.Image`` objects."""

	def on_paths(self, *args):
		"""Make textures from the images in ``paths``, and assign them at the
		same index in my ``texs`` as in my ``paths``.

		"""
		for i, path in enumerate(self.paths):
			if path in self.pathtexs:
				if (
					self.pathtexs[path] in self.texs
					and self.texs.index(self.pathtexs[path]) == i
				):
					continue
			else:
				try:
					self.pathimgs[path] = img = Image.load(
						resource_find(path), keep_data=True
					)
				except Exception:
					self.pathimgs[path] = img = Image.load(
						resource_find("atlas://rltiles/misc/floppy"),
						keep_data=True,
					)
				self.pathtexs[path] = img.texture
			if i == len(self.texs):
				self.texs.append(self.pathtexs[path])
			else:
				self.texs[i] = self.pathtexs[path]

	def clear(self):
		"""Clear paths, textures, rectangles"""
		self.paths = []
		super().clear()

	def insert(self, i, v):
		"""Insert a string to my paths"""
		if not isinstance(v, str):
			raise TypeError("Paths only")
		self.paths.insert(i, v)

	def __delitem__(self, i):
		"""Delete texture, rectangle, path"""
		super().__delitem__(i)
		del self.paths[i]

	def pop(self, i=-1):
		"""Delete and return a path"""
		r = self.paths[i]
		del self[i]
		return r


class TextureStackBatchWidget(Widget):
	"""Widget for efficiently drawing many TextureStacks

	Only add TextureStack or ImageStack widgets to this. Avoid adding
	any that are to be changed frequently.

	"""

	critical_props = ["texs", "offxs", "offys", "pos"]
	"""Properties that, when changed on my children, force a redraw."""

	def __init__(self, **kwargs):
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self._trigger_rebind_children = Clock.create_trigger(
			self.rebind_children
		)
		super(TextureStackBatchWidget, self).__init__(**kwargs)

	def on_parent(self, *args):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		if not hasattr(self, "_fbo"):
			with self.canvas:
				self._fbo = Fbo(size=self.size)
				self._fbo.add_reload_observer(self.redraw)
				self._translate = Translate(x=self.x, y=self.y)
				self._rectangle = Rectangle(
					texture=self._fbo.texture, size=self.size
				)
		self.rebind_children()

	def rebind_children(self, *args):
		child_by_uid = {}
		binds = {prop: self._trigger_redraw for prop in self.critical_props}
		for child in self.children:
			child_by_uid[child.uid] = child
			child.bind(**binds)
		if hasattr(self, "_old_children"):
			old_children = self._old_children
			for uid in set(old_children).difference(child_by_uid):
				old_children[uid].unbind(**binds)
		self.redraw()
		self._old_children = child_by_uid

	def redraw(self, *args):
		fbo = self._fbo
		fbo.bind()
		fbo.clear()
		fbo.clear_buffer()
		fbo.release()
		for child in self.children:
			assert child.canvas not in fbo.children
			fbo.add(child.canvas)

	def on_pos(self, *args):
		if not hasattr(self, "_translate"):
			return
		self._translate.x, self._translate.y = self.pos

	def on_size(self, *args):
		if not hasattr(self, "_rectangle"):
			return
		self._rectangle.size = self._fbo.size = self.size
		self.redraw()

	def add_widget(self, widget, index=0, canvas=None):
		if not isinstance(widget, TextureStack):
			raise TypeError("TextureStackBatch is only for TextureStack")
		if index == 0 or len(self.children) == 0:
			self.children.insert(0, widget)
		else:
			children = self.children
			if index >= len(children):
				index = len(children)

			children.insert(index, widget)
		widget.parent = self
		if hasattr(self, "_fbo"):
			self.rebind_children()

	def remove_widget(self, widget):
		if widget not in self.children:
			return
		self.children.remove(widget)
		widget.parent = None
		if hasattr(self, "_fbo"):
			self.rebind_children()


class TextureStackPlane(Widget):
	"""A widget full of Stacks, which are like TextureStack but more performant

	Stacks aren't technically :class:`Widget`s, though they have similar
	properties. This is because, when you have thousands of them, :class:`Widget`
	performs badly.

	"""

	data = ListProperty()
	selected = ObjectProperty(allownone=True)
	color_selected = ListProperty([0.0, 1.0, 1.0, 1.0])
	default_image_path = StringProperty("atlas://rltiles/base.atlas/unseen")

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._keys = []
		self._left_xs = np.array([])
		self._right_xs = np.array([])
		self._top_ys = np.array([])
		self._bot_ys = np.array([])
		self._instructions = {}
		self._stack_index = {}
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self._redraw_bind_data_uid = self.fbind("data", self._trigger_redraw)
		self._redraw_bind_size_uid = self.fbind("size", self._trigger_redraw)

	def on_parent(self, *_):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		with self.canvas:
			self._fbo = Fbo(size=self.size)
			self._push = PushMatrix()
			self._translate = Translate(x=self.x, y=self.y)
			self._rectangle = Rectangle(
				size=self.size, texture=self._fbo.texture
			)
			self._pop = PopMatrix()
		self.bind(pos=self._trigger_redraw, size=self._trigger_redraw)
		self._trigger_redraw()

	@mainthread
	def _add_datum_upd_fbo(self, **datum):
		name = datum["name"]
		texs = datum["textures"]
		x = datum["x"]
		y = datum["y"]
		fbo = self._fbo
		with fbo:
			instructions = self._instructions
			rects = []
			wide = datum.get("width", 0)
			tall = datum.get("height", 0)
			for tex in texs:
				if isinstance(tex, str):
					tex = Image.load(resource_find(tex)).texture
					w, h = tex.size
					if "width" not in datum and w > wide:
						wide = w
					if "height" not in datum and h > tall:
						tall = h
				rects.append(
					Rectangle(texture=tex, pos=(x, y), size=(wide, tall))
				)
			instructions[name] = {
				"rectangles": rects,
				"group": InstructionGroup(),
			}
			grp = instructions[name]["group"]
			for rect in rects:
				grp.add(rect)
			fbo.add(instructions[name]["group"])

	def add_datum(self, datum):
		name = datum["name"]
		if "pos" in datum:
			x, y = datum.pop("pos")
		else:
			x = datum["x"]
			y = datum["y"]
		if isinstance(x, float):
			x *= self.width
		if isinstance(y, float):
			y *= self.height
		left_xs = list(self._left_xs)
		right_xs = list(self._right_xs)
		top_ys = list(self._top_ys)
		bot_ys = list(self._bot_ys)
		left_xs.append(x)
		bot_ys.append(y)
		wide = datum.get("width", 0)
		tall = datum.get("height", 0)
		top_ys.append(y + tall)
		right_xs.append(x + wide)
		self._left_xs = np.array(left_xs)
		self._bot_ys = np.array(bot_ys)
		self._top_ys = np.array(top_ys)
		self._right_xs = np.array(right_xs)
		self._stack_index[name] = len(self._keys)
		self._keys.append(name)
		self.unbind_uid("data", self._redraw_bind_data_uid)
		self.data.append(datum)
		self._redraw_bind_data_uid = self.fbind("data", self._trigger_redraw)
		self._add_datum_upd_fbo(**datum)

	@mainthread
	def _remove_upd_fbo(self, name):
		if name not in self._instructions:
			return
		grp = self._instructions[name]["group"]
		grp.clear()
		fbo = self._fbo
		fbo.bind()
		fbo.remove(grp)
		fbo.ask_update()
		fbo.clear_buffer()
		fbo.release()
		self._rectangle.texture = fbo.texture
		del self._instructions[name]

	def remove(self, name_or_idx):
		def delarr(arr, i):
			if i == 0:
				return arr[1:]
			elif i == len(arr) - 1:
				return arr[:-1]
			else:
				return np.concatenate((arr[:i], arr[i + 1 :]))

		if name_or_idx in self._keys:
			idx = self._keys.index(name_or_idx)
			name = name_or_idx
		else:
			idx = name_or_idx
			name = self._keys[idx]
		stack_index = self._stack_index
		del stack_index[name]
		del self._keys[idx]
		for key in self._keys[idx:]:
			stack_index[key] -= 1
		self._left_xs = delarr(self._left_xs, idx)
		self._bot_ys = delarr(self._bot_ys, idx)
		self._top_ys = delarr(self._top_ys, idx)
		self._right_xs = delarr(self._right_xs, idx)
		self.unbind_uid("data", self._redraw_bind_data_uid)
		del self.data[idx]
		self._redraw_bind_data_uid = self.fbind("data", self._trigger_redraw)
		self._remove_upd_fbo(name)

	def redraw(self, *_):
		def get_rects(datum):
			width = datum.get("width", 0)
			height = datum.get("height", 0)
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			rects = []
			for texture in datum["textures"]:
				if isinstance(texture, str):
					try:
						texture = Image.load(resource_find(texture)).texture
					except Exception:
						texture = Image.load(self.default_image_path).texture
				w, h = texture.size
				if "width" in datum:
					w = width
				elif w > width:
					width = w
				if "height" in datum:
					h = height
				elif h > height:
					height = h
				assert w > 0 and h > 0
				rects.append(
					Rectangle(pos=(x, y), size=(w, h), texture=texture)
				)
			return rects

		def get_lines_and_colors(datum) -> dict:
			width = datum.get("width", 0)
			height = datum.get("height", 0)
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			right = x + width
			top = y + height
			instructions = {}
			colr = Color(rgba=color_selected)
			instructions["color0"] = colr
			line = Line(points=[x, y, right, y, right, top, x, top, x, y])
			instructions["line"] = line
			coler = Color(rgba=[1, 1, 1, 1])
			instructions["color1"] = coler
			return instructions

		if not hasattr(self, "_rectangle"):
			self._trigger_redraw()
			return
		if not self.data:
			if hasattr(self, "_fbo"):
				self._fbo.clear()
			return
		Logger.debug(
			f"TextureStackPlane: redrawing, with {self.selected} selected"
		)
		start_ts = monotonic()
		instructions = self._instructions
		stack_index = self._stack_index
		keys = list(self._keys)
		left_xs = list(self._left_xs)
		right_xs = list(self._right_xs)
		top_ys = list(self._top_ys)
		bot_ys = list(self._bot_ys)
		self_width = self.width
		self_height = self.height
		selected = self.selected
		color_selected = self.color_selected
		todo = []
		observed = set()
		for datum in self.data:
			name = datum["name"]
			observed.add(name)
			texs = datum["textures"]
			if isinstance(datum["x"], float):
				x = datum["x"] * self_width
			else:
				if not isinstance(datum["x"], int):
					raise TypeError("need int or float for pos")
				x = datum["x"]
			if isinstance(datum["y"], float):
				y = datum["y"] * self_height
			else:
				if not isinstance(datum["y"], int):
					raise TypeError("need int or float for pos")
				y = datum["y"]
			if name in stack_index:
				rects = get_rects(datum)
				if name == selected:
					insts = get_lines_and_colors(datum)
				else:
					insts = {}
				insts["rectangles"] = rects
				if name in instructions:
					insts["group"] = instructions[name]["group"]
				else:
					insts["group"] = InstructionGroup()
				todo.append(insts)
				instructions[name] = insts
				width = datum.get("width", 0)
				height = datum.get("height", 0)
				for texture, rect in zip(texs, rects):
					if isinstance(texture, str):
						try:
							texture = Image.load(
								resource_find(texture)
							).texture
						except Exception:
							texture = Image.load(
								self.default_image_path
							).texture
					w, h = texture.size
					if "width" in datum:
						w = width
					elif w > width:
						width = w
					if "height" in datum:
						h = height
					elif h > height:
						height = h
					rect.texture = texture
					assert w > 0 and h > 0
					rect.size = (w, h)
				idx = stack_index[name]
				right = x + width
				left_xs[idx] = x
				right_xs[idx] = right
				bot_ys[idx] = y
				top = y + height
				top_ys[idx] = top
			else:
				width = datum.get("width", 0)
				height = datum.get("height", 0)
				stack_index[name] = len(keys)
				keys.append(name)
				rects = get_rects(datum)
				grp = InstructionGroup()
				if "width" not in datum or "height" not in datum:
					width, height = rects[0].size
				right = x + width
				left_xs.append(x)
				right_xs.append(right)
				bot_ys.append(y)
				top = y + height
				top_ys.append(top)
				instructions[name] = insts = {
					"rectangles": rects,
					"group": grp,
				}
				if name == selected:
					insts.update(get_lines_and_colors(datum))
				todo.append(insts)
		unobserved = instructions.keys() - observed
		get_rid = []
		for gone in unobserved:
			get_rid.append(instructions.pop(gone))
		self._left_xs = np.array(left_xs)
		self._right_xs = np.array(right_xs)
		self._top_ys = np.array(top_ys)
		self._bot_ys = np.array(bot_ys)
		self._keys = keys
		self._fbo.bind()
		self._fbo.clear_buffer()
		self._fbo.release()
		self._fbo.size = self.size
		fbo = self._fbo
		for insts in todo:
			group = insts["group"]
			group.clear()
			for rect in insts["rectangles"]:
				group.add(rect)
			if "color0" in insts:
				group.add(insts["color0"])
				group.add(insts["line"])
				group.add(insts["color1"])
			if group not in fbo.children:
				fbo.add(group)
		for insts in get_rid:
			fbo.remove(insts["group"])
		self._rectangle.texture = fbo.texture
		self._translate.x, self._translate.y = self.pos
		self._rectangle.size = self._fbo.size
		Logger.debug(
			f"TextureStackPlane: redrawn in "
			f"{monotonic() - start_ts:,.2f} seconds"
		)

	def iter_collided_keys(self, x, y):
		"""Iterate over the keys of stacks that collide with the given point"""
		hits = (
			(self._left_xs <= x)
			& (self._bot_ys <= y)
			& (y <= self._top_ys)
			& (x <= self._right_xs)
		)
		return map(itemgetter(0), filter(itemgetter(1), zip(self._keys, hits)))


class Stack:
	__slots__ = ["board", "_name", "proxy", "__self__"]

	default_image_paths = ["atlas://rltiles/floor.atlas/floor-normal"]

	def __init__(self, **kwargs):
		self.board = kwargs["board"]
		if "proxy" in kwargs:
			self.proxy = kwargs["proxy"]
			self._name = kwargs["proxy"]["name"]
		elif "name" in kwargs:
			self._name = kwargs["name"]
		else:
			raise TypeError("Stacks need names")

	@property
	def paths(self):
		name = self.name
		plane = self._stack_plane
		datum = plane.data[plane._stack_index[name]]
		return datum["textures"]

	@paths.setter
	@mainthread
	def paths(self, v):
		name = self.name
		plane = self._stack_plane
		datum = plane.data[plane._stack_index[name]]
		plane.unbind_uid("data", plane._redraw_bind_data_uid)
		datum["textures"] = v
		insts = plane._instructions[name]
		rects = insts["rectangles"]
		group = insts["group"]
		for rect in rects:
			group.remove(rect)
		rects = insts["rectangles"] = []
		wide = datum.get("width", 0)
		tall = datum.get("height", 0)
		if v is None:
			v = self.default_image_paths
		for path in v:
			if not isinstance(path, str):
				raise TypeError("paths must be strings")
			tex = Image.load(path).texture
			w, h = tex.size
			if "width" not in datum and w > wide:
				wide = w
			if "height" not in datum and h > tall:
				tall = h
			rect = Rectangle(texture=tex, pos=self.pos, size=(wide, tall))
			rects.append(rect)
			group.add(rect)
		plane._redraw_bind_data_uid = plane.fbind(
			"data", plane._trigger_redraw
		)

	@property
	def selected(self):
		return self._stack_plane.selected == self.name

	@selected.setter
	@mainthread
	def selected(self, v: bool):
		stack_plane: TextureStackPlane = self._stack_plane
		name = self.name
		Logger.debug(f"Stack: {name} selected")
		insts = stack_plane._instructions[name]
		fbo = stack_plane._fbo
		fbo.bind()
		fbo.clear_buffer()
		if v:
			stack_plane.selected = name
			if "color0" in insts:
				Logger.debug(
					f"Stack: changing {name}'s color to {stack_plane.color_selected}"
				)
				insts["color0"].rgba = stack_plane.color_selected
			else:
				Logger.debug(
					f"Stack: creating Color(rgba={stack_plane.color_selected}) for {name}"
				)
				idx = stack_plane._stack_index[name]
				left = stack_plane._left_xs[idx]
				bot = stack_plane._bot_ys[idx]
				right = stack_plane._right_xs[idx]
				top = stack_plane._top_ys[idx]
				grp = insts["group"]
				insts["color0"] = Color(rgba=stack_plane.color_selected)
				grp.add(insts["color0"])
				insts["line"] = Line(
					points=[
						left,
						bot,
						right,
						bot,
						right,
						top,
						left,
						top,
						left,
						bot,
					]
				)
				grp.add(insts["line"])
				insts["color1"] = Color(rgba=[1.0, 1.0, 1.0, 1.0])
				grp.add(insts["color1"])
		else:
			if stack_plane.selected == self.name:
				stack_plane.selected = None
			if "color0" in insts:
				insts["color0"].rgba = [0.0, 0.0, 0.0, 0.0]
		fbo.release()

	@property
	def pos(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.name]
		return float(stack_plane._left_xs[idx]), float(
			stack_plane._bot_ys[idx]
		)

	@pos.setter
	@mainthread
	def pos(self, xy):
		x, y = xy
		stack_plane = self._stack_plane
		stack_plane.unbind_uid("data", stack_plane._redraw_bind_data_uid)
		name = self.name
		insts = stack_plane._instructions[name]
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		bot = stack_plane._bot_ys[idx]
		right = stack_plane._right_xs[idx]
		top = stack_plane._top_ys[idx]
		width = right - left
		height = top - bot
		r = x + width
		t = y + height
		stack_plane._left_xs[idx] = x
		stack_plane._bot_ys[idx] = y
		stack_plane._top_ys[idx] = t
		stack_plane._right_xs[idx] = r
		stack_plane._fbo.bind()
		stack_plane._fbo.clear_buffer()
		for rect in insts["rectangles"]:
			rect: Rectangle
			rect.pos = xy
			rect.flag_update()  # undocumented. sounds right?
		if "line" in insts:
			insts["line"].points = [x, y, r, y, r, t, x, t, x, y]
		stack_plane.data[idx]["pos"] = xy
		stack_plane._redraw_bind_data_uid = stack_plane.fbind(
			"data", stack_plane._trigger_redraw
		)
		stack_plane._fbo.release()

	@property
	def _stack_plane(self):
		return self.board.stack_plane

	@property
	def x(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.name]
		return float(stack_plane._left_xs[idx])

	@x.setter
	def x(self, x):
		self.pos = x, self.y

	@property
	def y(self):
		stack_plane = self._stack_plane
		idx = stack_plane._stack_index[self.name]
		return float(stack_plane._bot_ys[idx])

	@y.setter
	def y(self, y):
		self.pos = self.x, y

	@property
	def size(self):
		stack_plane = self._stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		bot = stack_plane._bot_ys[idx]
		right = stack_plane._right_xs[idx]
		top = stack_plane._top_ys[idx]
		return float(right - left), float(top - bot)

	@size.setter
	def size(self, wh):
		w, h = wh
		stack_plane = self._stack_plane
		stack_plane.unbind_uid("data", stack_plane._redraw_bind_data_uid)
		name = self.name
		insts = stack_plane._instructions[name]
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx] = x + w
		t = stack_plane._top_ys[idx] = y + h
		for rect in insts["rectangles"]:
			rect.size = wh
		if "line" in insts:
			insts["line"].points = [x, y, r, y, r, t, x, t, x, y]
		stack_plane.data[idx]["size"] = wh
		stack_plane._redraw_bind_data_uid = stack_plane.fbind(
			"data", stack_plane._trigger_redraw
		)

	@property
	def width(self):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		left = stack_plane._left_xs[idx]
		right = stack_plane._right_xs[idx]
		return float(right - left)

	@width.setter
	def width(self, w):
		self.size = self.height, w

	@property
	def height(self):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		top = stack_plane._top_ys[idx]
		bot = stack_plane._bot_ys[idx]
		return float(top - bot)

	@height.setter
	def height(self, h):
		self.size = self.width, h

	@property
	def center(self):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx]
		t = stack_plane._top_ys[idx]
		w = r - x
		h = t - y
		return float(x + w / 2), float(y + h / 2)

	@center.setter
	def center(self, c):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		y = stack_plane._bot_ys[idx]
		r = stack_plane._right_xs[idx]
		t = stack_plane._top_ys[idx]
		w = r - x
		h = t - y
		self.pos = c[0] - w / 2, c[1] - h / 2

	@property
	def center_x(self):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		r = stack_plane._right_xs[idx]
		w = r - x
		return float(x + w / 2)

	@center_x.setter
	def center_x(self, cx):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		r = stack_plane._right_xs[idx]
		w = r - x
		self.pos = cx - w / 2, self.y

	@property
	def center_y(self):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		t = stack_plane._top_ys[idx]
		h = t - y
		return float(y + h / 2)

	@center_y.setter
	def center_y(self, cy):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		t = stack_plane._top_ys[idx]
		h = t - y
		self.pos = self.x, cy - h / 2

	@property
	def top(self):
		stack_plane = self.board.stack_plane
		name = self.name
		if name not in stack_plane._stack_index:
			return 100
		idx = stack_plane._stack_index[name]
		return float(stack_plane._top_ys[idx])

	@top.setter
	def top(self, t):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		y = stack_plane._bot_ys[idx]
		stack_plane._top_ys[idx] = t
		h = t - y
		stack_plane._bot_ys[idx] = t - h
		self.pos = self.x, t - h

	@property
	def right(self):
		stack_plane = self.board.stack_plane
		name = self.name
		if name not in stack_plane._stack_index:
			return 100
		idx = stack_plane._stack_index[name]
		return float(stack_plane._right_xs[idx])

	@right.setter
	def right(self, r):
		stack_plane = self.board.stack_plane
		name = self.name
		idx = stack_plane._stack_index[name]
		x = stack_plane._left_xs[idx]
		stack_plane._right_xs[idx] = r
		w = r - x
		x = stack_plane._left_xs[idx] = r - w
		self.pos = x, self.y

	@property
	def name(self):
		return self._name

	def collide_point(self, x, y):
		pos = self.pos
		if x < pos[0] or y < pos[1]:
			return False
		w, h = self.size
		return x < pos[0] + w and y < pos[1] + h
